
#ifndef __VIDEO_SURFACE_OUTPUT_H
#define __VIDEO_SURFACE_OUTPUT_H

#include "pvmi_mio_control.h"
#include "pvmi_media_transfer.h"
#include "oscl_scheduler_ao.h"
#include "pvmi_media_io_observer.h"
#include "oscl_file_io.h"
#include "pvmi_config_and_capability.h"
#include "oscl_string_containers.h"
#include "pvmi_media_io_clock_extension.h"
#include "pvmf_media_msg_header.h"

#include <fcntl.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

// SurfaceFlinger
#include <ui/ISurface.h>

// interprocess shared memory support
#include <binder/MemoryBase.h>
#include <binder/MemoryHeapBase.h>

#include "cczoomrotation16.h"
#include "android_surface_output.h"

#include "hwdec_base_imp.h"
#include "dec_proxy_interface.h"
#include "hwdec_interface.h"

enum DecMediaTimeStatus
{
	DEC_MEDIA_ERROR = 0,
    	DEC_MEDIA_ONTIME,
    	DEC_MEDIA_LATE ,
    	DEC_MEDIA_EARLY
};

enum RenderAOState
{
	Render_uninitialized = 0,
	Render_initialized,
	Render_running,
	Render_paused,
	Render_quit
};

typedef void (*TCallback)(OsclAny *param);


/*
	Provide RenderAO with RenderProcInterface to deal with render stuff
*/
class RenderProcInterface
{
	public:
		virtual ~RenderProcInterface(){}
		
		DecMediaTimeStatus virtual RenderCheckTimestamp(int32 ealry,int32 late,PVMFTimestamp ts, uint32 &delta) = 0;
		void virtual RenderDataCallback(uint8 *data, uint32 size, OsclAny *context = NULL ) = 0;
		void virtual RenderOverNotificationCallback(OsclAny *context = NULL ) = 0;
		PVMFStatus virtual RenderTimeoutCallback(uint32 &callbackID, uint32 timeout, TCallback callback, OsclAny *context = NULL) = 0;
		PVMFStatus virtual RenderCancelTimeoutCallback( uint32 callbackID ) = 0;
};

/*
	Provide clock state change method
*/
class ClockStateUpdateCallbackInterface
{
	public :

		virtual ~ClockStateUpdateCallbackInterface(){}

		virtual PVMFStatus ClockStateChangedCallback(
			PVMFMediaClock::PVMFMediaClockState oldState, PVMFMediaClock::PVMFMediaClockState newState) = 0;
};

/*
	Active timing surport
*/
class VideoMIOActiveTiming :
		public PvmiClockExtensionInterface, 
		public PVMFMediaClockStateObserver ,
		public PVMFMediaClockNotificationsObs
{
	public:

		VideoMIOActiveTiming(ClockStateUpdateCallbackInterface *clockIf) :
			iClock(NULL),
			iClockNotificationsInf(NULL),
			iClockStateCallbackInf(clockIf),
			iClockState(PVMFMediaClock::STOPPED),
			iStartTime(0)
		{
			iTimeoutRequestQueue.reserve(2);
		}

		~VideoMIOActiveTiming()
		{
			if ( iClock && iClockNotificationsInf )
			{
				iClockNotificationsInf->RemoveClockStateObserver(*this);
				iClock->DestroyMediaClockNotificationsInterface(iClockNotificationsInf);
				iClockNotificationsInf = NULL;
			}

			cleanupRequestQueue();
		}

    		/*
    			PvmiClockExtensionInterface
    		*/
		OSCL_IMPORT_REF PVMFStatus SetClock(PVMFMediaClock* clockVal);

    		/*
    			PVMFMediaClockStateObserver
    		*/
		void ClockStateUpdated();

		/*
			Timestamp check
		*/
		DecMediaTimeStatus CheckTimestamp(int32 early, int32 late, PVMFTimestamp ts, uint32 & aDelta);

		/*
    			from PVInterface
    		*/
		OSCL_IMPORT_REF void addRef() {}
		OSCL_IMPORT_REF void removeRef() {}
		OSCL_IMPORT_REF bool queryInterface(const PVUuid& uuid, PVInterface*& iface);	

		/*
			PVMFMediaClockNotificationObs
		*/
		void ProcessCallBack(uint32 callBackID, PVTimeComparisonUtils::MediaTimeStatus aTimerAccuracy, uint32 aDelta,
                                     const OsclAny* aContextData, PVMFStatus aStatus);
		
		void NotificationsInterfaceDestroyed();

		
		PVMFMediaClock::PVMFMediaClockState clockState() { return iClockState; }

		PVMFStatus SetTimeout(uint32 &callbackID, uint32 timeout , TCallback callback , OsclAny *context );

		PVMFStatus CancelTimeout(uint32 callbackID);

		void cleanupRequestQueue();

		class TimeoutRequest
		{
			public :

				TimeoutRequest( uint32 id, TCallback callback , OsclAny* context):
					iCallbackID(id), iCallback(callback), iContext(context)
				{}
				
				uint32 iCallbackID;
				TCallback iCallback;
				OsclAny *iContext;
		};

	private:

		Oscl_Vector<TimeoutRequest, OsclMemAllocator> iTimeoutRequestQueue;
		PVMFMediaClock* iClock;
		PVMFMediaClockNotificationsInterface *iClockNotificationsInf;
		ClockStateUpdateCallbackInterface *iClockStateCallbackInf;
		PVMFMediaClock::PVMFMediaClockState iClockState;
		uint32 iStartTime;
};


/*
	RenderAO , active object dedicated to deal with render stuff
*/
class RenderAO : public OsclActiveObject
{
	public :
		RenderAO( RenderProcInterface *renderIf , uint32 bufferCount) :
			OsclActiveObject(OsclActiveObject::EPriorityNominal/*EPriorityNominal*/ , "VideoRender"),
			iBufferCount(bufferCount),
			iRenderIf(renderIf),
			iProcessIncomingRequest(false),
			iCurrentTimeoutID(0),
			iState(Render_uninitialized)
		{
			iRenderCommandQueue.reserve(bufferCount);
		}

		~RenderAO()
		{
			CleanupRenderCommandQueue();
		}
				
		class RenderCommand
		{
			public:

				RenderCommand( PVMFTimestamp ts, uint8 *data , uint32 len):
					iTs(ts),
					pData(data),
					iLen(len)
				{}
				
				~RenderCommand()
				{}
			
				PVMFTimestamp iTs;
				uint8 *pData;
				uint32 iLen;
		};

		void CleanupRenderCommandQueue();
		
		PVMFStatus sendCommand(PVMFTimestamp ts, uint8 *data , uint32 size);

		void Init();
		
		void Start();

		void Pause();

		void Exit();

		void Run();

		static void TimeoutCallback(OsclAny * param);

	private:

		uint32 iBufferCount;
		RenderProcInterface *iRenderIf;
		bool iProcessIncomingRequest;
		uint32 iCurrentTimeoutID;
		RenderAOState iState;
		Oscl_Vector<RenderCommand, OsclMemAllocator> iRenderCommandQueue;
};



class VideoSurfaceOutput :    public AndroidSurfaceOutput ,
		public DecodedDataProcessInterface ,
		public ClockStateUpdateCallbackInterface ,
		public RenderProcInterface
{

	public:

		VideoSurfaceOutput();
		~VideoSurfaceOutput();

		/*
			Overwritting some buffer managementl functions
		*/
    		PVMFStatus writeFrameBuf(
    			uint8* aData, 
    			uint32 aDataLen, 
    			const PvmiMediaXferHeader& data_header_info);

    		/*
    			Overwritting some componet processed functions
    		*/
    		PVMFCommandId Init(const OsclAny* aContext = NULL);


		PVMFCommandId Pause(const OsclAny * aContext = NULL);

		/*
			Called when stop
		*/
		PVMFCommandId Stop(const OsclAny * aContext = NULL);

		/*
			Overwritting reset function
		*/
		PVMFCommandId Reset(const OsclAny* aContext);

    		/*
    			Overwrite  virtuals from PvmiCapabilityAndConfig
    		*/
		PVMFCommandId QueryUUID(
			const PvmfMimeString & aMimeType, 
			Oscl_Vector < PVUuid, OsclMemAllocator > & aUuids, 
			bool aExactUuidsOnly, 
			const OsclAny * aContext);

		PVMFCommandId QueryInterface(
			const PVUuid & aUuid, 
			PVInterface * & aInterfacePtr, 
			const OsclAny * aContext);
		
    		PVMFStatus getParametersSync(
    			PvmiMIOSession aSession, 
    			PvmiKeyType aIdentifier,
            		PvmiKvp*& aParameters, 
            		int& num_parameter_elements, 
            		PvmiCapabilityContext aContext);

		void	setParametersSync(
			PvmiMIOSession aSession, 
			PvmiKvp * aParameters, 
			int num_elements, 
			PvmiKvp * & aRet_kvp);

		PVMFStatus verifyParametersSync (
			PvmiMIOSession aSession, 
			PvmiKvp* aParameters, 
			int num_elements);

		PVMFCommandId writeAsync(
			uint8 aFormatType, 
			int32 aFormatIndex, 
			uint8 * aData, 
			uint32 aDataLen, 
			const PvmiMediaXferHeader & data_header_info, 
			OsclAny * aContext);


		/* 
			Proxy Client related functions 
		*/
		void HandleCommandDone(
			PVMFStatus status, 
			TPVProxyMsgId cmdid);

		void HandleOneFrameDecoded(
			PVMFStatus status, 
			TPVProxyMsgId cmdid, 
			DEC_U32 param1, 
			DEC_U32 param2);

		static void BaseHandleCommandDone(
			DEC_HANDLETYPE hComponent, 
			PVMFStatus status, 
			TPVProxyMsgId cmdid);

		static void BaseHandleOneFrameDecoded(
			DEC_HANDLETYPE hComponent, 
			PVMFStatus status, 
			TPVProxyMsgId cmdid,
			DEC_U32 param1, 
			DEC_U32 param2);

		enum PendingCmdType
		{
			CMD_TYPE_CTRL = 0,
			CMD_TYPE_HEAD,
			CMD_TYPE_DATA,
		};

		class CommandPending
    		{
    			public:
        			CommandPending( 
						TPVProxyMsgId proxyId, 
						PVMFCommandId pvmfId, 
						const OsclAny* ctx , 
						PVMFTimestamp ts,
						PendingCmdType type)
            				: iProxyCmdId(proxyId), iPVMFCmdId(pvmfId) ,  iContext(ctx) , iTs(ts) , iType(type)
        			{}

				
        			TPVProxyMsgId iProxyCmdId;
        			PVMFCommandId iPVMFCmdId;
        			const OsclAny* iContext;
				PVMFTimestamp iTs;
				PendingCmdType iType;
    		};

		void CleanupCommandPendingQueue();
		
    		/*
    			DecodedDataProcessInterface,
    			This is useful when proxy interface disabled
		*/
		RET_STATUS OnePictureReadyCallback(DEC_UCHAR * pData, DEC_U32 nLen, DEC_PTR pParam);

		/*
			RenderProcInterface
		*/
		DecMediaTimeStatus RenderCheckTimestamp(int32 early,int32 late,PVMFTimestamp ts, uint32 &delta);
		
		void RenderDataCallback(uint8 *data, uint32 size, OsclAny * context = NULL);

		void RenderOverNotificationCallback(OsclAny * context = NULL);

		PVMFStatus RenderTimeoutCallback(uint32 &callbackID, uint32 timeout, TCallback callback, OsclAny *context = NULL);

		PVMFStatus RenderCancelTimeoutCallback(uint32 callbackID);

		/*
			Clock state change callback
		*/
		PVMFStatus ClockStateChangedCallback( 
			PVMFMediaClock::PVMFMediaClockState oldState, PVMFMediaClock::PVMFMediaClockState newState);

	private:

		PVMFStatus DecoderInitAndPrepare();

		void PictureResize(uint32 & resizeWidth, uint32 & resizeHeight);

		uint32 iLCDWidth;
		uint32 iLCDHeight;

		uint32 iSurfaceBufferWriteIndex;
		uint32 iSurfaceBufferReadIndex;

		/*
			This is only useful when proxy interface disabled
		*/
		PvmiMediaXferHeader *iCurrentMediaInfo;

		VideoMIOActiveTiming *iActiveTiming;
//		OsclSemaphore *iSurfaceBufferSem;
		RenderAO *iRenderAO;
		
		Oscl_Vector<CommandPending,OsclMemAllocator> iCommandPendingQueue;
		DEC_HANDLETYPE iDecComponentHandler;	
		uint8 *iTrackConfig;
		uint32 iTrackConfigSize;
		bool iHdrDecoded;
		bool iMIOActivated;
		
};

#endif /* __VIDEO_SURFACE_OUTPUT_H */

