/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
*
* Author 		: 	Garry fan
* Date		: 	2010 06 10
* Description 	:	MPEG4 Decoder interface
*
 */
 
#include "hantro8170_h264_decoder.h"
#include "hwdec_os_api.h"

DEC_HANDLETYPE H264DecoderFactory( 
	DecodedDataProcessInterface *dp , 
	DecProxyInterface *proxy , 
	DEC_PTR appContext , 
	CommandDoneCallback callback1,
	NotificationCallback callback2)
{
	Hantro8170H264Decoder *pH264 = new Hantro8170H264Decoder();

	if ( !pH264 )
	{
		HWDEC_ERR("No resource for allocating H264 Decoder");
		return NULL;
	}

	pH264->ConstructComponent(proxy , appContext, (DEC_PTR)pH264, callback1, callback2);
	pH264->SetDecodedDataProcInterface( dp, (DEC_PTR)( pH264->GetComponentHandle() ));

	return pH264->GetComponentHandle();
}


void H264DecoderDestructor( DEC_HANDLETYPE hComponent )
{
	Hantro8170H264Decoder *pH264 = 
		(Hantro8170H264Decoder *)(((DEC_COMPONENTTYPE *)hComponent)->decoderPrivateData);

	pH264->Deinit();

	delete pH264;
}


Hantro8170H264Decoder::Hantro8170H264Decoder( )
{
	iDecHandler = NULL;
	iFrameIndex = 0;

	oscl_memset(&iDecHeadInfo , 0 , sizeof(H264DecInfo));
}


Hantro8170H264Decoder::~Hantro8170H264Decoder()
{
}


RET_STATUS Hantro8170H264Decoder::DecSpecificInit()
{
	H264DecRet ret;

	ret = H264DecInit( &iDecHandler , 0 );

	if ( ret != H264DEC_OK )
	{
		HWDEC_ERR("MP4DecInit Error ret = %d" , ret);
		return Ret_ErrHardware;
	}

	return Ret_ErrNone;
}


RET_STATUS Hantro8170H264Decoder::DecSpecificDeinit()
{
	if ( iDecHandler )
		H264DecRelease(iDecHandler);

	iDecHandler = NULL;

	return Ret_ErrNone;
}


DEC_HANDLER Hantro8170H264Decoder::DecSpecificGetHandler()
{
	return (DEC_HANDLER)iDecHandler;
}


RET_STATUS Hantro8170H264Decoder::DecSpecificPrepare()
{
	/*
		Nothing to do here
	*/
	return Ret_ErrNone;
}


RET_STATUS Hantro8170H264Decoder::DecSpecificGetPPMode()
{
	return PP_PIPELINED_DEC_TYPE_H264;
}


RET_STATUS Hantro8170H264Decoder::GetVideoInfo(uint32 & oFormat, uint32 & displayWidth, uint32 & displayHeight, 
									uint32 & frameWidth, uint32 & frameHeight, uint32 & interlaced)
{
	oFormat = iDecHeadInfo.outputFormat;
	interlaced = 0;
	displayWidth = iDecHeadInfo.picWidth;
	displayHeight = iDecHeadInfo.picHeight;
	frameWidth = iDecHeadInfo.picWidth;
	frameHeight = iDecHeadInfo.picHeight;

	return Ret_ErrNone;
}


DEC_RET Hantro8170H264Decoder::VideoDecode(uint8 * aData, uint32 aBusAddr, uint32 aDataLen, 
									uint8 **outData, uint32 * outLen, uint32 * left, bool eos)
{
	H264DecInput 		decIn;
	H264DecOutput 	decOut;
	H264DecRet 		ret;

	decIn.picId = iFrameIndex;
	decIn.pStream = aData;
	decIn.streamBusAddress = aBusAddr;
	decIn.dataLen = aDataLen;

	do {

//		HWDEC_MSG("H264DecDecode <== pa:%08x va:%08x len:%d" , decIn.streamBusAddress , decIn.pStream , decIn.dataLen);
		ret = H264DecDecode( iDecHandler , &decIn , &decOut );
//		HWDEC_MSG("H264DecDecode ==> ret:%d left:%d", ret , decOut.dataLeft);
		
		switch(ret)
		{
			case H264DEC_HDRS_RDY:
			{
				H264DecRet ret;

				ret = H264DecGetInfo( iDecHandler , &iDecHeadInfo );

				if ( ret != H264DEC_OK )
					return DEC_RET_STREAM_ERR;

				break;
			}
		
			case H264DEC_PIC_DECODED:
			{
				H264DecPicture decPic;
				
				iFrameIndex++;

				while( H264DecNextPicture( iDecHandler, &decPic, eos ) == H264DEC_PIC_RDY )
				{

					/*
						If goes here , means we have successfully decoded one picture,
						time to set some indicator and return.
						We always give them the pOutputPicture as the outData pointer, the 
						caller knows where to get the actual data
					*/
					if ( outData)
						*outData = (uint8 *)decPic.pOutputPicture;

//					HWDEC_MSG("One pic decoded , picId = %d " , decPic.picId);

					/*
						One picture at a time
					*/
					break;

					/* 
						outLen ??? The caller knows the actual outLen , we dont care about it here  
					*/
				}

				break;
			}

			case H264DEC_STRM_PROCESSED:
				/* input stream processed but no picture ready */
//				HWDEC_MSG("Decoder stream processed");
				break;

			case H264DEC_STRM_ERROR:
				/* input stream processed but no picture ready */
				HWDEC_MSG("Decoder stream error");
				break;
				
			default:
				HWDEC_MSG("Decoder unknown retur ,ret = %d", ret);
				break;
		}

		*left = decOut.dataLeft;
		/*
			When a frame is decoded or heads ready , return 
		*/
		if ( ret == H264DEC_PIC_DECODED )
			return DEC_RET_PIC_DECODED;	
		
		else if ( ret == H264DEC_HDRS_RDY )
			return DEC_RET_HDR_READY;
		
		
		decIn.pStream = decOut.pStrmCurrPos;
		decIn.streamBusAddress = decOut.strmCurrBusAddress;
		decIn.dataLen = decOut.dataLeft;
		
	}while( decOut.dataLeft > 0  );

	return DEC_RET_STREAM_PROCESSED;
}



