/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
*
* Author 		: 	Garry fan
* Date		: 	2010 06 10
* Description 	:	MPEG4 Decoder interface
*
 */
 
#include "hantro8170_mpeg4_decoder.h"
#include "hwdec_os_api.h"
#include "hwdec_interface.h"

DEC_HANDLETYPE Mpeg4DecoderFactory( 
	DecodedDataProcessInterface *dp , 
	DecProxyInterface *proxy , 
	DEC_PTR appContext , 
	CommandDoneCallback callback1,
	NotificationCallback callback2)
{
	Hantro8170Mpeg4Decoder *pMpeg4 = new Hantro8170Mpeg4Decoder();

	if ( !pMpeg4 )
	{
		HWDEC_ERR("No resource for allocating H264 Decoder");
		return NULL;
	}

	pMpeg4->ConstructComponent(proxy , appContext, (DEC_PTR)pMpeg4, callback1, callback2);
	pMpeg4->SetDecodedDataProcInterface( dp, (DEC_PTR)(pMpeg4->GetComponentHandle()));

	return pMpeg4->GetComponentHandle();
}


void Mpeg4DecoderDestructor( DEC_HANDLETYPE hComponent )
{
	Hantro8170Mpeg4Decoder *pMpeg4 = 
		(Hantro8170Mpeg4Decoder *)(((DEC_COMPONENTTYPE *)hComponent)->decoderPrivateData);

	pMpeg4->Deinit();

	delete pMpeg4;
}


Hantro8170Mpeg4Decoder::Hantro8170Mpeg4Decoder()
{
	iDecHandler = NULL;
	iFrameIndex = 0;
	oscl_memset(&iDecHeadInfo , 0 , sizeof(MP4DecInfo));
}

Hantro8170Mpeg4Decoder::~Hantro8170Mpeg4Decoder()
{
}

RET_STATUS Hantro8170Mpeg4Decoder::DecSpecificInit()
{
	MP4DecRet ret;

	ret = MP4DecInit( &iDecHandler , 0 );

	if ( ret != MP4DEC_OK )
	{
		HWDEC_ERR("MP4DecInit Error ret = %d" , ret);
		return Ret_ErrHardware;
	}

	return Ret_ErrNone;
}


RET_STATUS Hantro8170Mpeg4Decoder::DecSpecificDeinit()
{
	if ( iDecHandler )
		MP4DecRelease(iDecHandler);

	iDecHandler = NULL;

	return Ret_ErrNone;
}


DEC_HANDLER Hantro8170Mpeg4Decoder::DecSpecificGetHandler()
{
	return (DEC_HANDLER)iDecHandler;
}


RET_STATUS Hantro8170Mpeg4Decoder::DecSpecificPrepare()
{
	/*
		Nothing to do here
	*/
	return Ret_ErrNone;
}


RET_STATUS Hantro8170Mpeg4Decoder::DecSpecificGetPPMode()
{
	return PP_PIPELINED_DEC_TYPE_MPEG4;
}


RET_STATUS Hantro8170Mpeg4Decoder::GetVideoInfo(uint32 & oFormat, uint32 & displayWidth, uint32 & displayHeight, 
									uint32 & frameWidth, uint32 & frameHeight, uint32 & interlaced)
{
	oFormat = iDecHeadInfo.outputFormat;
	interlaced = iDecHeadInfo.interlacedSequence;
	displayWidth = iDecHeadInfo.codedWidth;
	displayHeight = iDecHeadInfo.codedHeight;
	frameWidth = iDecHeadInfo.frameWidth;
	frameHeight = iDecHeadInfo.frameHeight;

	return Ret_ErrNone;
}


DEC_RET Hantro8170Mpeg4Decoder::VideoDecode(uint8 * aData, uint32 aBusAddr, uint32 aDataLen, 
									uint8 **outData, uint32 * outLen, uint32 * left, bool eos)
{
	MP4DecInput 		decIn;
	MP4DecOutput 	decOut;
	MP4DecRet 		ret;

	decIn.picId = iFrameIndex;
	decIn.pStream = aData;
	decIn.streamBusAddress = aBusAddr;
	decIn.dataLen = aDataLen;

	do {
		ret = MP4DecDecode( iDecHandler , &decIn , &decOut );

//		HWDEC_MSG("MP4DecDecode return with %d" , ret);

		switch(ret)
		{
			case MP4DEC_HDRS_RDY:
			{
				MP4DecRet ret;

				ret = MP4DecGetInfo( iDecHandler , &iDecHeadInfo );

				if ( ret != MP4DEC_OK )
					return DEC_RET_STREAM_ERR;
				
				break;
			}
		
			case MP4DEC_PIC_DECODED:
			{
				MP4DecPicture decPic;
				int secondField = 0;
				
				iFrameIndex++;

				while( MP4DecNextPicture( iDecHandler, &decPic, eos ) == MP4DEC_PIC_RDY )
				{
					if ((decPic.fieldPicture && secondField) ||!decPic.fieldPicture) 
					{
/*					
						HWDEC_MSG("AtmelMpeg4Decoder: "
								 "decoded picture %d, mpeg4 timestamp: %d:%d:%d:%d (%d)\n",
						 		iFrameIndex,
						 		decPic.timeCode.hours, decPic.timeCode.minutes,
						 		decPic.timeCode.seconds, decPic.timeCode.timeIncr,
						 		decPic.timeCode.timeRes);
*/
						if (decPic.fieldPicture)
							secondField = 0;

						/*
							If goes here , means we have successfully decoded one picture,
							time to set some indicator and return.
							We always give them the pOutputPicture as the outData pointer, the 
							caller knows where to get the actual data
						*/
						if ( outData)
							*outData = (uint8 *)decPic.pOutputPicture;

						/* 
							outLen ??? The caller knows the actual outLen , we dont care about it here  
						*/
						break;
					}
					else if ( decPic.fieldPicture )
						secondField = 1;
				}

				break;
			}

			case MP4DEC_STRM_PROCESSED:
				/* input stream processed but no picture ready */
//				HWDEC_MSG("Decoder stream processed");
				break;

			case MP4DEC_STRM_ERROR:
				/* input stream processed but no picture ready */
				HWDEC_MSG("Decoder stream error");
				break;

			case MP4DEC_VOS_END:
				/* end of stream */
				HWDEC_MSG("Decoder vos end");
				decOut.dataLeft = 0;
				break;
				
			default:
				HWDEC_MSG("Decoder unknown retur ,ret = %d", ret);
				break;
		}


		*left = decOut.dataLeft;

		/*
			When a frame is decoded or heads ready , return 
		*/
		if ( ret == MP4DEC_PIC_DECODED )
			return DEC_RET_PIC_DECODED;	
	
		if ( ret == MP4DEC_HDRS_RDY )
			return DEC_RET_HDR_READY;
		
		decIn.pStream = decOut.pStrmCurrPos;
		decIn.streamBusAddress = decOut.strmCurrBusAddress;
		decIn.dataLen = decOut.dataLeft;
		
	}while( decOut.dataLeft > 0 );

	return DEC_RET_STREAM_PROCESSED;
}



