/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
*
* Author 		: 	Garry fan
* Date		: 	2010 06 10
* Description 	:	Decoder base implementation
*
* Revision	:	2.0		Add proxy interface and async method
* Revision	:	1.1		Add work thread implementation
 */

 
#include "hwdec_os_api.h"
#include "hwdec_base_imp.h"
#include "dec_proxy_interface.h"
#include "oscl_mem_basic_functions.h"
#include "oscl_tickcount.h"

#define DUMP_PP_CONFIG

/*
	DecComponentBase constructor
	dp: Decoded data process interface
	proxy: the proxy interface we are going to rely on if asked to use proxy , not using proxy if assigned null.
	appContext: when using proxy interface , this contain the application context that will communicate 
			    with proxy server
*/
DecComponentBase::DecComponentBase()
{
	iDpInf = NULL;
	iDpParam = NULL;

	pp = NULL;
	LBAllocator = NULL;
	inputBuffer = outputBuffer = LBF_HANDLER_INVALID;	

	iOutputBufferBasePA = 0;
	iOutputBufferBaseVA = NULL;
	iOutputBufferOffset = NULL;
	iOutputBufferNum = 0;
	iOutputBufferIndex = 0;
	iOutputBufferSem = NULL;
	iOutputBufferUpdate = true;
	iCurrentBufferPA = 0;
	iCurrentBufferVA = NULL;
	iCurrentBufferSize = 0;
	
	iPeerConnected = false;

	iCropWidth = 0;
	iCropHeight = 0;
	iCropX = 0;
	iCropY = 0;
	iOutputWidth = 0;
	iOutputHeight = 0;
	iRotation = PP_ROTATION_NONE;
	iInDataFormat = 0;
	iDisplayWidth = 0;
	iDisplayHeight = 0;
	iFrameWidth = 0;
	iFrameHeight = 0;
	iInterlaced = 0;
	iDecOutFormat = 0;
	iImgOutFormat = 0;
}

/*
	fill the DEC_COMPONENTTYPE member's value.
	point to different interfaces according to request
*/
void DecComponentBase::ConstructComponent(DEC_PTR pProxy,
	DEC_PTR appContext, DEC_PTR decoderContext , 
	CommandDoneCallback callback1,NotificationCallback callback2)
{
	iComponent.componentPrivateData = this;
	iComponent.applicationPrivateData = appContext;
	iComponent.decoderPrivateData = decoderContext;

	iProxyApp = pProxy;
	
	/*	Proxy interface enabled */
	if ( iProxyApp ) 
	{
		iComponent.DEC_SetOutputImage = DecComponentBase::BaseComponentProxySetOutputImage;
		iComponent.DEC_SetRatation = DecComponentBase::BaseComponentProxySetInputRotation;
		iComponent.DEC_SetOutputCrop = DecComponentBase::BaseComponentProxySetOutputCrop;
		iComponent.DEC_Init = DecComponentBase::BaseComponentProxyInit;
		iComponent.DEC_Prepare = DecComponentBase::BaseComponentProxyPrepare;
		iComponent.DEC_Decode = DecComponentBase::BaseComponentProxyDecode;
		iComponent.DEC_Deinit = DecComponentBase::BaseComponentProxyDeinit;
		iComponent.DEC_GetHeaderInfo = DecComponentBase::BaseComponentProxyGetHeaderInfo;
		iComponent.DEC_GetCommandId = DecComponentBase::BaseComponentProxyGetCommandId;
		iComponent.DEC_ReturnBuffer = DecComponentBase::BaseComponentReturnBuffer;

		/*
			Proxy interface server end function pointer assignment
		*/
		((DecProxyInterface *)iProxyApp)->ComponentSetOutputImage = DecComponentBase::BaseComponentSetOutputImage;
		((DecProxyInterface *)iProxyApp)->ComponentSetInputRotation = DecComponentBase::BaseComponentSetInputRotation;
		((DecProxyInterface *)iProxyApp)->ComponentSetOutputCrop = DecComponentBase::BaseComponentSetOutputCrop;
		((DecProxyInterface *)iProxyApp)->ComponentInit = DecComponentBase::BaseComponentInit;
		((DecProxyInterface *)iProxyApp)->ComponentPrepare = DecComponentBase::BaseComponentPrepare;
		((DecProxyInterface *)iProxyApp)->ComponentDecode = DecComponentBase::BaseComponentDecode;
		((DecProxyInterface *)iProxyApp)->ComponentDeinit = DecComponentBase::BaseComponentDeinit;
		((DecProxyInterface *)iProxyApp)->ComponentGetHeaderInfo = DecComponentBase::BaseComponentGetHeaderInfo;
		((DecProxyInterface *)iProxyApp)->HandleCommandDoneCallback = callback1;
		((DecProxyInterface *)iProxyApp)->HandleOneFrameDecodedCallback = callback2;
	}
	else	/* no proxy interface used */
	{
		iComponent.DEC_SetOutputImage = DecComponentBase::BaseComponentSetOutputImage;
		iComponent.DEC_SetRatation = DecComponentBase::BaseComponentSetInputRotation;
		iComponent.DEC_SetOutputCrop = DecComponentBase::BaseComponentSetOutputCrop;
		iComponent.DEC_Init = DecComponentBase::BaseComponentInit;
		iComponent.DEC_Prepare = DecComponentBase::BaseComponentPrepare;
		iComponent.DEC_Decode = DecComponentBase::BaseComponentDecode;
		iComponent.DEC_Deinit = DecComponentBase::BaseComponentDeinit;
		iComponent.DEC_GetHeaderInfo = DecComponentBase::BaseComponentGetHeaderInfo;
		iComponent.DEC_GetCommandId = DecComponentBase::BaseComponentGetCommandId;
		iComponent.DEC_ReturnBuffer = DecComponentBase::BaseComponentReturnBuffer;
	}
}


/*
	interface without proxy entry point implementation
	STATIC functions cannot use class member directly , we should send the class pointer as 
	parameter to those static functions
*/
RET_STATUS DecComponentBase::BaseComponentSetOutputImage(
	DEC_HANDLETYPE hComponent, DEC_U32 nWidth, 
	DEC_U32 nHeight, DEC_U32 nOutputFormat)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	return pComponent->SetOutputImage( nWidth , nHeight, nOutputFormat);
}


RET_STATUS DecComponentBase::BaseComponentSetInputRotation(
	DEC_HANDLETYPE hComponent, DEC_U32 nRotation)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	return pComponent->SetInputRotation( nRotation);
}


RET_STATUS DecComponentBase::BaseComponentSetOutputCrop(
	DEC_HANDLETYPE hComponent, DEC_U32 nLeftTopX, 
	DEC_U32 nLeftTopY, DEC_U32 nWidth, DEC_U32 nHeight)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	return pComponent->SetOutputCrop( nLeftTopX, nLeftTopY, nWidth, nHeight);
}


RET_STATUS DecComponentBase::BaseComponentInit(
	DEC_HANDLETYPE hComponent)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	return pComponent->Init();
}


RET_STATUS DecComponentBase::BaseComponentPrepare(
	DEC_HANDLETYPE hComponent)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	return pComponent->Prepare();
}


RET_STATUS DecComponentBase::BaseComponentDecode(
	DEC_HANDLETYPE hComponent, DEC_UCHAR * pInputData, DEC_U32 nInputDataLen, DEC_BOOL bEos)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	return pComponent->Decode( pInputData ,nInputDataLen,bEos);
}


RET_STATUS DecComponentBase::BaseComponentDeinit(
	DEC_HANDLETYPE hComponent)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	pComponent->Deinit();

	return Ret_ErrNone;
}


RET_STATUS DecComponentBase::BaseComponentGetHeaderInfo(
	DEC_HANDLETYPE hComponent, DEC_U32 * pFrameWidth, DEC_U32 * pFrameHeight, 
	DEC_U32 * pDisplayWidth, DEC_U32 * pDisplayHeight, DEC_U32 * pDecOutFormat)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	return pComponent->GetHeaderInfo( pFrameWidth, pFrameHeight, 
								pDisplayWidth, pDisplayHeight, pDecOutFormat);
}

TPVProxyId DecComponentBase::BaseComponentGetCommandId(DEC_HANDLETYPE hComponent)
{
	/* Interface without proxy doesnt support this command */
	return -1;
}

/*
	Proxy interface entry point implementation
*/
RET_STATUS DecComponentBase::BaseComponentProxySetOutputImage(
	DEC_HANDLETYPE hComponent, DEC_U32 nWidth, 
	DEC_U32 nHeight, DEC_U32 nOutputFormat)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxySetOutputImage(hComponent , nWidth , nHeight, nOutputFormat);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxySetInputRotation(
	DEC_HANDLETYPE hComponent, DEC_U32 nRotation)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxySetInputRotation( hComponent, nRotation);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxySetOutputCrop(
	DEC_HANDLETYPE hComponent, DEC_U32 nLeftTopX, DEC_U32 nLeftTopY, DEC_U32 nWidth, DEC_U32 nHeight)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxySetOutputCrop( hComponent, nLeftTopX, nLeftTopY, nWidth, nHeight);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxyInit(
	DEC_HANDLETYPE hComponent)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxyInit(hComponent);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxyPrepare(
	DEC_HANDLETYPE hComponent)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxyPrepare(hComponent);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxyDecode(
	DEC_HANDLETYPE hComponent, DEC_UCHAR * pInputData, DEC_U32 nInputDataLen, DEC_BOOL bEos)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}
	
	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxyDecode(hComponent,pInputData ,nInputDataLen,bEos);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxyDeinit(
	DEC_HANDLETYPE hComponent)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxyDeinit(hComponent);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


RET_STATUS DecComponentBase::BaseComponentProxyGetHeaderInfo(
	DEC_HANDLETYPE hComponent, DEC_U32 * pFrameWidth, DEC_U32 * pFrameHeight, 
	DEC_U32 * pDisplayWidth, DEC_U32 * pDisplayHeight, DEC_U32 * pDecOutFormat)
{
	RET_STATUS ret;
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	ret = ((DecProxyInterface *)(pComponent->iProxyApp))->ProxyGetHeaderInfo(hComponent,pFrameWidth, pFrameHeight, 
												pDisplayWidth, pDisplayHeight, pDecOutFormat);

	if ( ret == Ret_ErrWait )
		pComponent->iCurrentAsyncCommandId = 
			((DecProxyInterface *)(pComponent->iProxyApp))->GetCurrentAsyncCommandId();

	return ret;
}


TPVProxyId DecComponentBase::BaseComponentProxyGetCommandId(DEC_HANDLETYPE hComponent)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return -1;
	}

	return pComponent->GetAsyncCommandId();
}


RET_STATUS DecComponentBase::BaseComponentReturnBuffer(DEC_HANDLETYPE hComponent)
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
	{
		HWDEC_ERR("Invalid component pointer!");
		return Ret_ErrInvalidArg;
	}

	return pComponent->OutputBufferRetrun();
}

RET_STATUS DecComponentBase::OutputBufferOpen(uint32 basePA , uint8 *baseVA, uint32 bufferSize , uint32 bufferNum )
{
	OSCL_ASSERT(basePA);
	OSCL_ASSERT(baseVA);

	uint32 i = 0;
	
	iOutputBufferBasePA = basePA;
	iOutputBufferBaseVA = baseVA;
	iOutputBufferOffset = (uint32 *)oscl_malloc(sizeof(uint32)*bufferNum);

	if ( !iOutputBufferOffset )
	{
		HWDEC_ERR("Cannot allocate memory for iOutputBufferOffset");
		return Ret_ErrNoMemory;
	}

	iOutputBufferSem = new OsclSemaphore();

	if ( !iOutputBufferSem )
	{
		oscl_free(iOutputBufferOffset);
		iOutputBufferOffset = NULL;
		HWDEC_ERR("Cannot allocaate memory for iOutputBufferSem");
		
		return Ret_ErrNoMemory;
	}

	iOutputBufferSem->Create(bufferNum);
	
	for( i = 0; i < bufferNum; i++ )
		iOutputBufferOffset[i] = bufferSize*i;

	iOutputBufferNum = bufferNum;
	iOutputBufferIndex = 0;

	return Ret_ErrNone;
}


void DecComponentBase::OutputBufferClose()
{
	if ( iOutputBufferOffset )
	{
		oscl_free(iOutputBufferOffset);
		iOutputBufferOffset = NULL;
	}

	if ( iOutputBufferSem )
	{
		iOutputBufferSem->Close();
		delete iOutputBufferSem;
		iOutputBufferSem = NULL;
	}

	iOutputBufferBasePA = 0;
	iOutputBufferBaseVA = NULL;
	iOutputBufferNum = 0;
	iOutputBufferIndex = 0;
}


uint32 DecComponentBase::OutputBufferIndexInc()
{
	uint32 newIndex = iOutputBufferIndex;

	if ( ++newIndex == iOutputBufferNum )
		newIndex = 0;

	return newIndex;
}


uint32 DecComponentBase::OutputBufferIndexDec()
{
	uint32 newIndex = iOutputBufferIndex;

	if ( newIndex == 0 )
		newIndex = iOutputBufferNum - 1;
	else
		newIndex--;

	return newIndex;
}


RET_STATUS DecComponentBase::OutputBufferGetNext( uint32 &pa, uint8 *&va, uint32 &size )
{
	/*
		Wait if no free buffer left
	*/
	if (iOutputBufferSem->Wait() != OsclProcStatus::SUCCESS_ERROR)
	{
		HWDEC_ERR("OutputBufferSem wait error!");
        	return Ret_ErrGeneral;
	}

	pa = iOutputBufferBasePA + iOutputBufferOffset[iOutputBufferIndex];
	va = iOutputBufferBaseVA + iOutputBufferOffset[iOutputBufferIndex];
	size = iOutputBufferOffset[1];

	iOutputBufferIndex = OutputBufferIndexInc();

	return Ret_ErrNone;
}

/*
	This is always called by the consumer 
*/
RET_STATUS DecComponentBase::OutputBufferRetrun()
{
	if ( iOutputBufferSem->Signal() != OsclProcStatus::SUCCESS_ERROR )
		return Ret_ErrGeneral;

	return Ret_ErrNone;
}

/*
	Native method implementation
*/
RET_STATUS DecComponentBase::SetOutputImage(DEC_U32 nWidth, DEC_U32 nHeight, 
											DEC_U32 nOutputFormat)
{
	if ( nOutputFormat != V_OUT_YUV 
			&& nOutputFormat != V_OUT_RGB15 
				&& nOutputFormat != V_OUT_RGB16 
					&& nOutputFormat != V_OUT_RGB32 )
		return Ret_ErrInvalidArg;

	iOutputWidth = nWidth;
	iOutputHeight = nHeight;
	iImgOutFormat = nOutputFormat;

	return Ret_ErrNone;
}


RET_STATUS DecComponentBase::SetOutputCrop(DEC_U32 nTopLeftX, DEC_U32 nTopLeftY, 
											DEC_U32 nWidth, DEC_U32 nHeight)
{
	/* Be clear of what you're doing */
	iCropX = nTopLeftX;
	iCropY = nTopLeftY;
	iCropWidth = nWidth;
	iCropHeight = nHeight;

	return Ret_ErrNone;
}


RET_STATUS DecComponentBase::SetInputRotation(DEC_U32 rotation)
{
	if ( rotation != PP_ROTATION_NONE 
		&& rotation != PP_ROTATION_HOR_FLIP 
			&& rotation != PP_ROTATION_VER_FLIP 
				&& rotation != PP_ROTATION_LEFT_90
					&& rotation != PP_ROTATION_RIGHT_90 
						&& rotation != PP_ROTATION_180)
		iRotation = PP_ROTATION_NONE;
	else
		iRotation = rotation;

	return Ret_ErrNone;
}


/*
	Init function ,get part of the resource ready
*/
RET_STATUS DecComponentBase::Init()
{
	if ( DecSpecificInit() != Ret_ErrNone )
	{
		HWDEC_ERR("Dec Hardware specific init error");
		return Ret_ErrHardware;
	}

	LBAllocator = new LinearBufferAllocator( VIDEO_DEFAULT_BUFFER_NUM , 
									VIDEO_DEFAULT_BUFFER_GROW );

	/* This should never happen */
	if ( !LBAllocator)
	{
		HWDEC_ERR(" new LinearBufferAllocator error ");
		DecSpecificDeinit();
		return Ret_ErrHardware;
	}

	iState = DEC_STAT_INITIALIZED;
	return Ret_ErrNone;	
}


/* 
	There is a strict order when calling those apis , and prepare sould be called 
	after all the parameters have been properly set, when error occured , caller should
	be responsible for calling deinit to remove all the resources outside this function ,
	it's supposed to be called once in a decoder lifetime
*/
RET_STATUS DecComponentBase::Prepare()
{
	if ( iState != DEC_STAT_INITIALIZED )
	{
		HWDEC_ERR(" Prepare invalid state ");
		return Ret_ErrInvalidState;
	}

	if ( DecSpecificPrepare() != Ret_ErrNone )
	{
		HWDEC_ERR("DecSpecificPrepare err!");
		return Ret_ErrGeneral;
	}

	/* 
		Let's assume that parameters have been set
	*/
	if ( iRotation != PP_ROTATION_NONE 
		|| iCropHeight != 0
			|| iCropWidth != 0 
				|| iImgOutFormat != V_OUT_YUV )
	{
		DEC_HANDLER decHandler;
		DEC_U32 ppMode;
		int i = 0;
		
		/* 
			Need support from post processor 
		*/
		pp = new HWPostProcessor();

		if ( !pp )
		{
			HWDEC_ERR("Cannot allocate resources for AtmelPostProcessor Object");
			return Ret_ErrNoMemory;
		}

		/* 
			Init hardware post processor 
		*/
		if ( pp->Init() != Ret_ErrNone )
		{
			HWDEC_ERR("Prepare error : pp->Init()");
			return Ret_ErrGeneral;
		}

		decHandler = DecSpecificGetHandler();

		if ( !decHandler )
		{
			HWDEC_ERR("Cannot get decoder handler");
			return Ret_ErrGeneral;
		}

		ppMode = DecSpecificGetPPMode();
		
		/* 
			Combine decoder and post processor 
		*/
		if ( pp->Connect( decHandler , ppMode ) != Ret_ErrNone )
		{
			HWDEC_ERR("Prepare error : pp->Connect");
			return Ret_ErrGeneral;
		}
		else
			iPeerConnected = true;

		/* 
			Allocate buffer for post processor 
		*/
		if ( LBAllocator->Allocate( VIDEO_OUTPUT_BUFFER_SIZE*VIDEO_OUTPUT_BUFFER_NUM, outputBuffer ) != Ret_ErrNone )
		{
			HWDEC_ERR("Prepare error : LBAllocator->Allocate() for outputbuffer");
			return Ret_ErrGeneral;
		}

		if ( OutputBufferOpen( LBAllocator->getBufferBussAddr(outputBuffer),
							LBAllocator->getBufferVirtualAddr(outputBuffer),
							VIDEO_OUTPUT_BUFFER_SIZE , 
							VIDEO_OUTPUT_BUFFER_NUM) != Ret_ErrNone )
		{
			HWDEC_ERR("Prepare error : cannot open output buffer");
			return Ret_ErrGeneral;
		}
		
		pp->SetConfig( PP_OUTIMG_HEIGHT , iOutputHeight);
		pp->SetConfig( PP_OUTIMG_WIDTH , iOutputWidth);
		pp->SetConfig( PP_OUTIMG_FORMAT , iImgOutFormat);
		
		/* 
			We dont set chrom pa here , since this value should be computed later 
		*/
	
#ifdef PP_CROP_SUPPORT
		pp->SetConfig( PP_CROP_X , iCropX );
		pp->SetConfig( PP_CROP_Y , iCropY );
		pp->SetConfig( PP_CROP_HEIGHT , iCropHeight );
		pp->SetConfig( PP_CROP_WIDTH , iCropWidth);
#endif

#ifdef PP_ROTATION_SUPPORT
		pp->SetConfig( PP_ROTA_ROTATION , iRotation );
#endif

	}


	/* 
		Allocate input buffer here 
	*/
	if ( LBAllocator->Allocate( VIDEO_DEFAULT_BUFFER_SIZE , inputBuffer ) != Ret_ErrNone )
	{
		HWDEC_ERR("Prepare error : LBAllocator->Allocate() for inputbuffer");
		return Ret_ErrNoMemory;
	}

	iState = DEC_STAT_PREPARED;
	return Ret_ErrNone;
}

/*
	This is the decode warpper layer function 
	It will keep decoding until all the input data has been consumed
	During the process , when a picture is decoded it will call the render callback function to handle it
	Both multi-thread app and single thread app will come to it finally
*/
RET_STATUS DecComponentBase::DecodeInternal(
	DEC_UCHAR * pVa, DEC_U32 nPa, DEC_U32 nSize, DEC_BOOL bEos)
{
	DEC_U32 consumed;
	DEC_U32 left;
	DEC_RET ret;
	DEC_UCHAR *pOutData;
	uint32 outLen;

	DEC_UCHAR *dataVA = pVa;
	DEC_U32 dataPA = nPa;
	DEC_U32 dataSize = nSize;

//#define DEC_PERFORMANCE_TEST

#ifdef DEC_PERFORMANCE_TEST
		uint32 tickCnt1 , tickCnt2 , msec;
		tickCnt1 = OsclTickCount::TickCount();	
#endif

	/*
		We need to get a new clean output buffer to contain the decoded data
	*/
	if ( iOutputBufferUpdate && pp )
	{
		if ( OutputBufferGetNext( iCurrentBufferPA , iCurrentBufferVA, iCurrentBufferSize ) != Ret_ErrNone )
		{
			HWDEC_ERR("Cannot get next free output buffer!");
			return Ret_ErrNoResources;
		}

//		HWDEC_ERR("Update -> pa = 0x%08x , va = 0x%08x , size = %d", 
//						iCurrentBufferPA, iCurrentBufferVA , iCurrentBufferSize);

		pp->SetOutputBufferAddrAndUpdate(iCurrentBufferPA);
		iOutputBufferUpdate = false;
	}

	while( dataSize > 0 )
	{
		ret = VideoDecode( dataVA , dataPA , dataSize , &pOutData, &outLen, &left , bEos);

		switch( ret )
		{
			case DEC_RET_HDR_READY:
			{
				GetVideoInfo( iDecOutFormat , iDisplayWidth , iDisplayHeight , 
										iFrameWidth , iFrameHeight , iInterlaced );
				if ( pp )
				{
					uint32 reqSize;
					RET_STATUS ret;
				
					pp->SetConfig( PP_INIMG_FORMAT , iDecOutFormat );
					pp->SetConfig( PP_INIMG_WIDTH , iFrameWidth );
					pp->SetConfig( PP_INIMG_HEIGHT , iFrameHeight );
					pp->SetConfig( PP_OUTIMG_DEINTERLACE , iInterlaced );
					
					/* 
						Tell the post processor to update parameters 
					*/
					ret =  pp->Updateconfig( iCurrentBufferPA , iCurrentBufferSize , reqSize );

					if ( ret == Ret_ErrResize )
					{
						/*
							May be we can extend to resize the buffer in the furue
						*/
						HWDEC_ERR("The buffer provided too small");
						return Ret_ErrGeneral;
					}
					else if ( ret != Ret_ErrNone )
					{
						HWDEC_ERR("Pp Update Config error");
						return Ret_ErrGeneral;
					}
					
			#ifdef DUMP_PP_CONFIG
					pp->DumpConfig();
			#endif
				}
					
				iState = DEC_STAT_HEADER_READY;
				break;
			}

			case DEC_RET_PIC_DECODED:
			{
				if ( pp )
				{
					outLen = pp->getCurrentPicSize();
					pOutData = (uint8 *)iCurrentBufferVA;
				}
				else
				{
					outLen = iOutputWidth*iOutputHeight*3/2;
					/* outData has been assigned by VideoDecode */
				}

				/*
					iDpParam should point to the decoder component handle
				*/
				if ( iDpInf->OnePictureReadyCallback( pOutData , outLen , iDpParam ) 
						!= Ret_ErrNone)
					HWDEC_ERR("Picture Ready callback error");
				else
					iOutputBufferUpdate = true;

				if ( iState == DEC_STAT_HEADER_READY )
					iState == DEC_STAT_DECODING;
	
				break;
			}

			case DEC_RET_STREAM_NOT_SUPPORTED:
			case DEC_RET_STREAM_ERR:
			{
				iState = DEC_STAT_ERR;
				return Ret_ErrGeneral;
			}
				
			case DEC_RET_STREAM_PROCESSED:
//				HWDEC_ERR("DEC STREAM PROCESSED");
				break;

			default:
				HWDEC_ERR("Unkown state  = %d" , ret);
				break;
					
		}
			/*
				Update data source information
			*/
		consumed = dataSize - left;
		dataVA += consumed;
		dataPA += consumed;
		dataSize = left;		
	}/* Keep decoding until the buffer data all consumed */

#ifdef DEC_PERFORMANCE_TEST
		tickCnt2 = OsclTickCount::TickCount();
		HWDEC_MSG("Decode one frame take <== : tick = %u" , tickCnt2 - tickCnt1);
#endif

	return Ret_ErrNone;
}



/*
	This is the core decode function .
	It will not return until all the input data has been consumed or critical error occured.
	It takes into acount the input stream case , in which the input data is not a single frame,
	but a partial fram or multi-frame , after each frame is successfully decoded , the registered
	Data process callback function will be invodeked to deal with the decode media data.
*/
RET_STATUS DecComponentBase::Decode(
	DEC_UCHAR * pInputData, DEC_U32 nInputDataLen, DEC_BOOL bEos)
{
	DEC_U32 		sourceLeft , decLeft;
	DEC_U32		cpySize;
	DEC_UCHAR	*pSource;

	DEC_UCHAR 	*dataVA;
	DEC_U32 		dataPA;
	DEC_U32	 	dataSize;

	DEC_RET		ret;

	DEC_UCHAR *const	bufferVA = (uint8 *)LBAllocator->getBufferVirtualAddr(inputBuffer);
	const DEC_U32	bufferPA = LBAllocator->getBufferBussAddr(inputBuffer);
	const DEC_U32	bufferSize = LBAllocator->getBufferSize(inputBuffer);

	/*
		Make judgement on state
	*/
	if ( iState != DEC_STAT_PREPARED &&
		iState != DEC_STAT_HEADER_READY &&
			iState != DEC_STAT_DECODING )
	{
		HWDEC_ERR("Decode invalid state = %d", iState);
		return Ret_ErrInvalidState;
	}

	pSource= pInputData;
	sourceLeft = nInputDataLen;

	do{
		/*
			Determine how many bytes need to copy to buffer
		*/
		cpySize = (sourceLeft > bufferSize)?bufferSize:sourceLeft;
		oscl_memcpy( bufferVA , pSource , cpySize );
		
		dataVA = bufferVA;
		dataPA = bufferPA;
		dataSize = cpySize;

		sourceLeft -= cpySize;
		pSource += cpySize;

		if ( bEos && sourceLeft == 0 )
			DecodeInternal( dataVA, dataPA, dataSize, bEos);
		else
			DecodeInternal( dataVA , dataPA , dataSize, NULL);
		
	}while( sourceLeft > 0 ); /* Keep decoding until all the inputbuffer consumed */

	return Ret_ErrNone;
}



/*
	Reset all the parameters and hardware resources
*/
void DecComponentBase::Deinit()
{	
	if ( iState == DEC_STAT_IDLE )
		return ;

	if ( LBAllocator )
	{
		if ( inputBuffer != LBF_HANDLER_INVALID )
		{
			LBAllocator->DeAllocate( inputBuffer );
			inputBuffer = LBF_HANDLER_INVALID;
		}	

		if ( outputBuffer != LBF_HANDLER_INVALID )
		{
			LBAllocator->DeAllocate( outputBuffer );
			outputBuffer = LBF_HANDLER_INVALID;
		}
		
		LBAllocator->DeInit();
		delete LBAllocator;
		LBAllocator = NULL;
	}

	if ( pp )
	{
		DEC_HANDLER decHandler;

		decHandler = DecSpecificGetHandler();
	
		if ( iPeerConnected && decHandler )
			pp->Disconnect(decHandler);

		pp->Deinit();
		delete pp;
		pp = NULL;

		iPeerConnected = false;
	}

	DecSpecificDeinit();
	iState = DEC_STAT_IDLE;

}


/*
	For upper MIO component to retrive video specific information
	This  function return success only after video header has been decoded
*/
RET_STATUS DecComponentBase::GetHeaderInfo(
	DEC_U32 * pFrameWidth, DEC_U32 * pFrameHeight, DEC_U32 * pDisplayWidth, 
	DEC_U32 * pDisplayHeight, DEC_U32 * pDecOutFormat)
{
	if ( iState != DEC_STAT_HEADER_READY && 
					iState != DEC_STAT_DECODING )
		return Ret_ErrInvalidState;

	*pFrameWidth = iFrameWidth;
	*pFrameHeight = iFrameHeight;
	*pDisplayWidth = iDisplayWidth;
	*pDisplayHeight = iDisplayHeight;
	*pDecOutFormat = iDecOutFormat;

	return Ret_ErrNone;
}



