


#include "hwdec_factory.h"
#include "hantro8170_mpeg4_decoder.h"
#include "hantro8170_h264_decoder.h"

#define MAX_NAME_LEN		64
#define MAX_SUPPORT_NUM	10

#define H264_DEC_SUPPORT
#define MPEG4_DEC_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////

typedef DEC_HANDLETYPE 
	(*decoderFactoryFunc)( DecodedDataProcessInterface *, DecProxyInterface * ,
							DEC_PTR , CommandDoneCallback, NotificationCallback);

typedef void (*decoderDestructorFunc)( DEC_HANDLETYPE );

typedef struct
{
	char	name[MAX_NAME_LEN];
	DEC_INPUT_FORMAT inFmt[MAX_SUPPORT_NUM];
	DEC_OUTPUT_FORMAT outFmt[MAX_SUPPORT_NUM];
	decoderFactoryFunc Create;
	decoderDestructorFunc Destruct;
}DecoderEntry;

static DecoderEntry *gCurrentDecoder = NULL;

////////////////////////////////////////////////////////////////////////////////////////

extern DEC_HANDLETYPE H264DecoderFactory(
			DecodedDataProcessInterface * dp, 
			DecProxyInterface * proxy, 
			DEC_PTR appContext, 
			CommandDoneCallback callback1,
			NotificationCallback callback2);

extern void H264DecoderDestructor(DEC_HANDLETYPE hComponent);


extern DEC_HANDLETYPE Mpeg4DecoderFactory(
			DecodedDataProcessInterface * dp, 
			DecProxyInterface * proxy, 
			DEC_PTR appContext, 
			CommandDoneCallback callback1,
			NotificationCallback callback2);


extern void Mpeg4DecoderDestructor(DEC_HANDLETYPE hComponent);


static DecoderEntry DecoderList[] = 
{

#ifdef MPEG4_DEC_SUPPORT
	/* MEPG4 , H263 DECODER */
	{ 	
		"Hantro8170_Mpeg4Decoder" ,
		{ 	IN_FMT_H263 , IN_FMT_MPEG4		} ,
		{ 	OUT_FMT_RGB555 , OUT_FMT_RGB565 , OUT_FMT_RGB32	},
		Mpeg4DecoderFactory,
		Mpeg4DecoderDestructor
	} ,
#endif

#ifdef H264_DEC_SUPPORT
	/* H264 DECODER */
	{
		"Hantro8170_H264Decoder" ,
		{	IN_FMT_H264						},
		{	OUT_FMT_RGB555 , OUT_FMT_RGB565 , OUT_FMT_RGB32	},
		H264DecoderFactory,
		H264DecoderDestructor
	},
#endif
	/* Add new decoder entry here */
};


static DecoderEntry *FindMatchDecoder(DEC_INPUT_FORMAT inFmt , DEC_OUTPUT_FORMAT outFmt)
{
	uint32	i , j ,k;
	DecoderEntry *pDecoder;

	for ( i = 0; i < sizeof(DecoderList) /sizeof(DecoderEntry) ; i++  )
	{
		pDecoder = &DecoderList[i];

		for( j = 0; j < MAX_SUPPORT_NUM; j++ )
		{
			if ( pDecoder->inFmt[j] == inFmt )
			{
				for ( k = 0; k < MAX_SUPPORT_NUM; k++ )
				{
					if ( pDecoder->outFmt[k] == outFmt )
					{
						HWDEC_MSG("Found matching decoder ==> %s", pDecoder->name);

						/* 
							Oh , we finally got the right entry , so uneasy!! 
						*/
						return pDecoder;
					}
				}

				/* 
					Goes here means we dont have any entries match the request out format 
					so , we dont have to check other input format
				*/
				break;
			}
		}
	}

	/*
		Nothing found matches the request
	*/
	return NULL;
}

DEC_HANDLETYPE CreateDecoder(DecodedDataProcessInterface *dp,
	DecProxyInterface *proxy, DEC_PTR pAppData, 
	CommandDoneCallback callback1, NotificationCallback callback2,
        DEC_INPUT_FORMAT inFmt , DEC_OUTPUT_FORMAT outFmt)
{
	DecoderEntry *pDecoder = FindMatchDecoder( inFmt, outFmt );
	DEC_HANDLETYPE handle = NULL;

	if ( pDecoder )
	{
		handle = pDecoder->Create(dp,proxy, pAppData,callback1,callback2);

		if ( handle )
			gCurrentDecoder = pDecoder;
	}
	
	return handle;
}

void DestructDecoder(DEC_HANDLETYPE hComponent)
{
	gCurrentDecoder->Destruct(hComponent);
	gCurrentDecoder = NULL;
}


RET_STATUS HWDEC_GetHandle( DEC_HANDLETYPE* pHandle, DecodedDataProcessInterface *dp,
        DEC_PTR pAppData, DEC_BOOL useProxy, 
        CommandDoneCallback callback1, NotificationCallback callback2 , 
        DEC_INPUT_FORMAT inFmt , DEC_OUTPUT_FORMAT outFmt)
{
	if ( useProxy)
	{
		DecProxyInterface *pProxy = new DecProxyInterface();
		DecodedDataProcessInterface *proxyDp = (DecodedDataProcessInterface *)pProxy;

		if ( pProxy->GetMemPoolPtr() == NULL )
			return Ret_ErrNoMemory;
		
		if (! pProxy->Start() )
			return Ret_ErrNoResources;

		/*
			If use proxy , so we use the proxy defined DecodedDataProcessInterface
		*/
		if ( pProxy->ProxyGetHandle( pHandle, proxyDp , pProxy, pAppData , callback1, callback2, inFmt, outFmt) == Ret_ErrNone )
			return Ret_ErrNone;
		else
		{
			/*
				Delete the proxy interface and return with an error
			*/
			pProxy->Exit();
			delete pProxy;
			
			return Ret_ErrGeneral;
		}
	}
	/*
		None proxied interface , we use the DecodedDataProcessInterface provided by APP
	*/
	else
	{
		*pHandle = CreateDecoder( dp,  NULL , pAppData, NULL ,NULL, inFmt, outFmt);

		if ( *pHandle == NULL )
			return Ret_ErrGeneral;
		
		return Ret_ErrNone;
	}
}



RET_STATUS HWDEC_FreeHandle( DEC_HANDLETYPE hComponent )
{
	DecComponentBase *pComponent = 
		(DecComponentBase *)((DEC_COMPONENTTYPE *)hComponent)->componentPrivateData;

	if ( !pComponent )
		return Ret_ErrInvalidArg;

	DecProxyInterface *iProxy = (DecProxyInterface *)(pComponent->GetProxyInterface());

	if ( iProxy )
	{
		RET_STATUS ret;

		ret = iProxy->ProxyFreeHandle(hComponent);

		if ( ret != Ret_ErrNone )
		{
			HWDEC_ERR("HWDEC_FreeHandle proxyFreeHandle error!");
		}
		
		iProxy->Exit();

		delete iProxy;
		
		return ret;
	}
	else
	{
		DestructDecoder(hComponent);
	}

	return Ret_ErrNone;
}


