/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */
#include "hwdec_interface.h"
#include "hwdec_post_proc.h"
#include "hwdec_os_api.h"
#include "ppapi.h"

HWPostProcessor::HWPostProcessor()
{
	ppHandler = NULL;
	decHandler = NULL;

	iConnected = false;
	iOutImgSize = 0;
}


HWPostProcessor::~HWPostProcessor()
{
	Deinit();
}


RET_STATUS HWPostProcessor::Init()
{
	if ( PPInit( &ppHandler ) != PP_OK )
		return Ret_ErrHardware;

	return Ret_ErrNone;
}

RET_STATUS HWPostProcessor::Deinit()
{
	if ( iConnected && decHandler )
	{
		PPDecCombinedModeDisable( ppHandler , decHandler);
		decHandler = NULL;
		iConnected = false;
	}
	
	if ( ppHandler )
	{
		PPRelease( ppHandler );
		ppHandler = NULL;
	}

	return Ret_ErrNone;
}

RET_STATUS HWPostProcessor::Connect(DEC_HANDLER peerHandler , uint32 decType)
{
	if ( PPDecCombinedModeEnable( ppHandler , peerHandler , decType) != PP_OK )
		return Ret_ErrHardware;

	decHandler = peerHandler;
	iConnected = true;

	return Ret_ErrNone;
}


RET_STATUS HWPostProcessor::Disconnect(DEC_HANDLER peerHandler )
{
	if ( !iConnected && !decHandler )
		return Ret_ErrNone;

	if ( PPDecCombinedModeDisable( ppHandler , peerHandler ) != PP_OK )
		return Ret_ErrHardware;

	return Ret_ErrNone;
}


void HWPostProcessor::InputRotationParamCheck( uint32 &rotation , uint32 &mWidth , uint32 &mHeight )
{
	switch ( rotation )
	{
		case PP_ROTATION_NONE:
		case PP_ROTATION_180:
			break;
			
		case PP_ROTATION_RIGHT_90:
		case PP_ROTATION_LEFT_90:
		{
			uint32 tmp;
			
			tmp = mWidth;
			mWidth = mHeight;
			mHeight = tmp;
			break;
		}	

		default:
			rotation = PP_ROTATION_NONE;
			break;
	}
}


void HWPostProcessor::OutputImgParamCheck( uint32 width , uint32 height , uint32 &outWidth , uint32 &outHeight )
{
	if ( outWidth == 0 && outHeight == 0 )
		return ;

	/* We can not upscale width and downscale height */ 
	if( ( outWidth > width) && ( outHeight < height) ) 
		outWidth  = width;
		
	/* We can not downscale width and upscale height */ 
	if( ( outWidth < width) && ( outHeight > height) ) 
		outHeight = height;

	if ( outWidth < 16 )
		outWidth = 16;

	if ( outWidth > 1280 )
		outWidth = 1280;

	if ( outHeight < 16 )
		outHeight = 16;

	if ( outHeight > 720 )
		outHeight = 720;

	if ( outWidth > (3*width) )
		outWidth = 3*width;

	if ( outHeight > (3*height - 2) )
		outHeight = 3*height -2;

	outWidth = (outWidth/8)*8;
	outHeight = (outHeight/2)*2;	
}


RET_STATUS HWPostProcessor::SetConfig(uint32 id, uint32 val)
{
	char param[ sizeof(paramKvp) + sizeof(uint32) ];
	paramKvp *pKvp = (paramKvp *)param;

	pKvp->key = id;
	pKvp->type = KVP_INT;
	HWDEC_MEM_CPY( pKvp->val , &val , sizeof(val));

	return configData.setConfig( pKvp , sizeof(paramKvp) + sizeof(uint32) );
}


RET_STATUS HWPostProcessor::GetConfig( uint32 id , uint32 &val )
{
	uint32 len = sizeof(paramKvp) + sizeof(uint32);
	char param[len];
	paramKvp *pKvp = (paramKvp *)param;

	if ( configData.getConfig( pKvp , &len) != Ret_ErrNone )
		return Ret_ErrGeneral;

	/* This is not necessary , we'll get what we'v set before */
	if ( len != sizeof(paramKvp) + sizeof(uint32) )
		return Ret_ErrGeneral;

	val = (uint32)pKvp->val[0];

	return Ret_ErrNone;
}

uint32 HWPostProcessor::getCurrentPicSize()
{
	return iOutImgSize;
}

void HWPostProcessor::DumpConfig()
{
	PPConfig pcfg;
	
#define PINIMG 	pcfg.ppInImg
#define POUTIMG 	pcfg.ppOutImg
#define PROTA	pcfg.ppInRotation
	

	if ( PPGetConfig( ppHandler , &pcfg ) != PP_OK )
		return ;

	HWDEC_MSG("Post processor config info :\n");
	HWDEC_MSG("ppInImg : w = %d , h = %d , f = %d" , 
				PINIMG.width , PINIMG.height , PINIMG.pixFormat);
	HWDEC_MSG("ppOutImg: w = %d , h = %d , f = %d , ba = 0x%08x , bca = 0x%08x",
				POUTIMG.width , POUTIMG.height , POUTIMG.pixFormat , 
				POUTIMG.bufferBusAddr , POUTIMG.bufferChromaBusAddr);
	HWDEC_MSG("rotation : en = %d " , PROTA.rotation);

} 

bool HWPostProcessor::SetOutputBufferAddrAndUpdate(uint32 pa)
{
	PPConfig pcfg;
	PPResult ppret;

	if ( PPGetConfig( ppHandler , &pcfg ) != PP_OK )
		return false;

	pcfg.ppOutImg.bufferBusAddr = pa;
	pcfg.ppOutImg.bufferChromaBusAddr = pa + pcfg.ppOutImg.width*pcfg.ppOutImg.height;
	
//	HWDEC_MSG("[PP]:out-> 0x%08x , chrom-> 0x%08x", 
//					pcfg.ppOutImg.bufferBusAddr , pcfg.ppOutImg.bufferChromaBusAddr);

	if ( ( ppret = PPSetConfig( ppHandler , &pcfg ) )  != PP_OK ) 
	{
		HWDEC_ERR("cannot set. PP settings (err=%d)", ppret );
		return false;
	}

	return true;
}

RET_STATUS HWPostProcessor::Updateconfig( uint32 pa , uint32 bufferSize , uint32 &reqSize)
{
	PPConfig ppconfig;
	PPResult ppret;
	paramKvp *tmpKvp = NULL;
	uint32 tmpLen = 0;
	uint32 mWidth , mHeight;
	uint32 outputImgSize = 0;
	uint32 tmpOutFormat;

	ppret = PPGetConfig( ppHandler , &ppconfig);

	if ( ppret != PP_OK )
	{
		HWDEC_ERR(" Get post processor configuration error! ");
		return false;
	}

	tmpLen = sizeof(paramKvp) + sizeof(uint32);
	tmpKvp = (paramKvp *)HWDEC_MALLOC( tmpLen );

	if ( !tmpKvp )
	{
		HWDEC_ERR(" No resources for tmp kvp ");
		return false;
	}

	/* For all the operation below , kvp type is alwawys KVP_INT */
	tmpKvp->type = KVP_INT;

	/* For INIMG parameters */
	tmpKvp->key = PP_INIMG_WIDTH;
	
	/* Cannot get configuration properly */
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;
	
	ppconfig.ppInImg.width = *(uint32 *)(tmpKvp->val);
	ppconfig.ppInImg.width = (ppconfig.ppInImg.width + 7) & ~7;
	mWidth = ppconfig.ppInImg.width;

	tmpKvp->key = PP_INIMG_HEIGHT;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	ppconfig.ppInImg.height = *(uint32 *)(tmpKvp->val);
	ppconfig.ppInImg.height = (ppconfig.ppInImg.height + 7) & ~7;
	mHeight = ppconfig.ppInImg.height;
	

	tmpKvp->key = PP_INIMG_FORMAT;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;
	
	ppconfig.ppInImg.pixFormat = *(uint32 *)(tmpKvp->val);
	ppconfig.ppInImg.videoRange = 1; /* fixed value for now */


	/* Dont support crop right now */
	ppconfig.ppInCrop.enable = 0;

#ifdef PP_ROTATION_SUPPORT
	/* For rotation parameters */
	tmpKvp->key = PP_ROTA_ROTATION;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	ppconfig.ppInRotation.rotation = *(uint32 *)(tmpKvp->val);
	InputRotationParamCheck( ppconfig.ppInRotation.rotation, mWidth, mHeight);
#else
	ppconfig.ppInRotation.rotation = PP_ROTATION_NONE;
#endif	


	/* For output parameters  */
	tmpKvp->key = PP_OUTIMG_HEIGHT;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	ppconfig.ppOutImg.height = *(uint32 *)(tmpKvp->val);


	tmpKvp->key = PP_OUTIMG_WIDTH;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	ppconfig.ppOutImg.width = *(uint32 *)(tmpKvp->val);

	OutputImgParamCheck( mWidth , mHeight,
						ppconfig.ppOutImg.width,
						ppconfig.ppOutImg.height);

	/* Set the modified output dimension back */
	tmpKvp->key = PP_OUTIMG_WIDTH;
	HWDEC_MEM_CPY( tmpKvp->val , &ppconfig.ppOutImg.width , sizeof(uint32));
	if ( configData.setConfig( tmpKvp , sizeof(tmpKvp) + sizeof(uint32)) != Ret_ErrNone ) 
		goto err_out;

	tmpKvp->key = PP_OUTIMG_HEIGHT;
	HWDEC_MEM_CPY( tmpKvp->val , &ppconfig.ppOutImg.height , sizeof(uint32));
	if ( configData.setConfig( tmpKvp , sizeof(tmpKvp) + sizeof(uint32)) != Ret_ErrNone )
		goto err_out;

	tmpKvp->key = PP_OUTIMG_FORMAT;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	tmpOutFormat = *(uint32 *)(tmpKvp->val);

	switch( tmpOutFormat )
	{
		case V_OUT_YUV:	/* YUV should not appear here theoriticaly */
			ppconfig.ppOutImg.pixFormat = PP_PIX_FMT_YCBCR_4_2_0_SEMIPLANAR;
			iOutImgSize = ( ppconfig.ppOutImg.width*ppconfig.ppOutImg.height )*3/2;
			break;
			
		case V_OUT_RGB15:
			ppconfig.ppOutImg.pixFormat = PP_PIX_FMT_RGB16_5_5_5;
			iOutImgSize = ( ppconfig.ppOutImg.width*ppconfig.ppOutImg.height )*2;
			break;
			
		case V_OUT_RGB16:
			ppconfig.ppOutImg.pixFormat = PP_PIX_FMT_RGB16_5_6_5;
			iOutImgSize = ( ppconfig.ppOutImg.width*ppconfig.ppOutImg.height )*2;
			break;
			
		case V_OUT_RGB32:
			ppconfig.ppOutImg.pixFormat = PP_PIX_FMT_RGB32;
			iOutImgSize = ( ppconfig.ppOutImg.width*ppconfig.ppOutImg.height )*4;
			break;

		default:
			HWDEC_ERR(" Ilegal Output Image Format! ");
			iOutImgSize = 0;
			goto err_out;
			break;
	}

	/* We need to make sure there is enought room to contain the decoded image */
	if ( bufferSize < iOutImgSize )
	{
		reqSize = iOutImgSize;
		HWDEC_ERR( " Updateconfig : Buffer not big enough, need to resize " );
		return Ret_ErrResize;
	}

	/* Set buffer parameters for pp */
	ppconfig.ppOutImg.bufferBusAddr = pa;
	ppconfig.ppOutImg.bufferChromaBusAddr = pa + ppconfig.ppOutImg.width*ppconfig.ppOutImg.height;


	/* For out interlace */
	tmpKvp->key = PP_OUTIMG_DEINTERLACE;
	if ( configData.getConfig( tmpKvp , &tmpLen ) != Ret_ErrNone )
		goto err_out;

	ppconfig.ppOutDeinterlace.enable = *(uint32 *)(tmpKvp->val);


	/* For out RGB param */
	ppconfig.ppOutRgb.alpha = 255;

	if ( ( ppret = PPSetConfig( ppHandler , &ppconfig ) )  != PP_OK ) 
	{
		HWDEC_ERR("cannot set. PP settings (err=%d)", ppret );
		goto err_out;
	}

	if( tmpKvp )
		HWDEC_FREE(tmpKvp);
	
	return Ret_ErrNone;

err_out:

	if ( tmpKvp )
		HWDEC_FREE( tmpKvp );

	return Ret_ErrGeneral;
	
}


