
#include "linear_malloc.h"
#include "hwdec_os_api.h"
#include <errno.h>

#ifdef DEFAULT_BUFFER_THRESH
#undef DEFAULT_BUFFER_THRESH
#endif

#define DEV_MEMALLOC_PATH		"/dev/memalloc"

#define DEFAULT_BUFFER_THRESH 	4
#define DEFAULT_BUFFER_GROW		( (DEFAULT_BUFFER_THRESH) >> 1)

LinearBufferAllocator::LinearBufferAllocator( uint32 numThresh , uint32 grow):
	iBufferNum(0),
	iBufferNumGrow(DEFAULT_BUFFER_GROW),
	iBufferNumThresh(DEFAULT_BUFFER_THRESH),
	initCompleted(false), 
	iMemAllocFd(-1),
	BufferPool(NULL)
{
	iBufferNumThresh = numThresh;
	iBufferNumGrow = grow;
}

RET_STATUS LinearBufferAllocator::Init()
{
	RET_STATUS status = Ret_ErrNone;

	if ( !BufferPool )
	{
		if ( !(BufferPool = (LinearBufferCtrl **)HWDEC_MALLOC(sizeof(LinearBufferCtrl *)*iBufferNumThresh) ) )
		{
			HWDEC_ERR("[LinearBufferAllocator]:Allocate memory for BufferPool error!\n");
			status = Ret_ErrNoMemory;
			goto err_out;
		}
	}

	HWDEC_MEM_SET( BufferPool , 0 , sizeof(LinearBufferCtrl *)*iBufferNumThresh);

	if ( iMemAllocFd < 0 )
	{
		if ( ( iMemAllocFd = HWDEC_OPEN(DEV_MEMALLOC_PATH, O_RDWR | O_SYNC) ) < 0 )
		{
			HWDEC_ERR("%s open err!\n" , DEV_MEMALLOC_PATH );
			status = Ret_ErrOpenDevice;
			goto err_out;
		}
	}

	initCompleted = true;
	return status;

err_out:
	DeInit();
	return status;

}


void LinearBufferAllocator::DeInit()
{
	uint32 i;

	if ( BufferPool )
	{
		for( i = 0; i < iBufferNumThresh; i++ )
		{
			if ( BufferPool[i] )
			{
				DeAllocate( (LBUF_HANDLER)i );
			}
		}

		HWDEC_FREE( BufferPool );
		BufferPool = NULL;
	}

	if ( iMemAllocFd > 0 )
	{
		HWDEC_ERR("Close memalloc fd = %d" , iMemAllocFd);
		HWDEC_CLOSE( iMemAllocFd );
		iMemAllocFd = -1;
	}

	initCompleted = false;
}



LinearBufferCtrl * LinearBufferAllocator::getLinearBufferCtrlByHandler(LBUF_HANDLER handler)
{
	uint32 index = (uint32)handler;

	if ( index >= iBufferNumThresh ||initCompleted == false )
		return NULL;

	return BufferPool[index];
}



RET_STATUS LinearBufferAllocator::Allocate( uint32 size , LBUF_HANDLER &handler)
{
	uint32 i;
	int pgsize = getpagesize();
	MemallocParams params;

	if ( initCompleted == false )
	{
		if ( Init() != Ret_ErrNone )
			return Ret_ErrGeneral;
	}
	
	/* Get a free entry for the buffer pool */
	for( i = 0; i < iBufferNumThresh; i++ )
	{
		if ( !BufferPool[i] )
		{
			BufferPool[i] = (LinearBufferCtrl *)HWDEC_MALLOC(sizeof(LinearBufferCtrl));

			if ( !BufferPool[i] )
			{
				HWDEC_ERR("[LinearBufferAllocator]: Insufficient resources for BufferPool\n");
				return Ret_ErrNoMemory;
			}
			else
			{
				HWDEC_MEM_SET( BufferPool[i] , 0 , sizeof(LinearBufferCtrl) );
				break;
			}
		}
	}

	/* All the entries are in use , have to enlarge the bufferpool */
	if ( i == iBufferNumThresh )
	{
		LinearBufferCtrl **NewBufferPool;

		NewBufferPool = (LinearBufferCtrl **)HWDEC_MALLOC( sizeof(LinearBufferCtrl *)*( iBufferNumThresh + iBufferNumGrow ) );

		if ( !NewBufferPool )
		{
			HWDEC_ERR("[LinearBufferAllocator]: Cannot allocate memory for buffer pool growth!\n");
			return Ret_ErrNoMemory;
		}

		HWDEC_MEM_SET( NewBufferPool , 0 , sizeof(LinearBufferCtrl*)*(iBufferNumThresh + iBufferNumGrow) );
		HWDEC_MEM_CPY( NewBufferPool , BufferPool , sizeof(LinearBufferCtrl)*iBufferNumThresh );
		HWDEC_FREE(BufferPool);

		BufferPool = NewBufferPool;
		iBufferNumThresh += iBufferNumGrow;

		/* Assign a linearBufferCtrl to this entry */
		BufferPool[i] = (LinearBufferCtrl *)HWDEC_MALLOC(sizeof(LinearBufferCtrl));

		if ( !BufferPool[i] )
		{
			HWDEC_ERR("Allocate memory for LinearBufferCtrl error!\n");
			return Ret_ErrNoMemory;
		}

	}

	/* When LinearBufferCtrl allocated , we should increase the num */
	iBufferNum++;

	/* Linear buffer size should be pgsize aligned */
	size = (size + (pgsize - 1)) & (~(pgsize - 1));
	params.size = size;
	HWDEC_IOCTL( iMemAllocFd , MEMALLOC_IOCXGETBUFFER, &params);

	if ( params.busAddress == 0 )
	{
		HWDEC_ERR("Cannot allocate linear buffer from mem device , size = %08x\n" , size);
		DeAllocate( (LBUF_HANDLER) i);
		return Ret_ErrNoResources;
	}

	BufferPool[i]->bussAddr = params.busAddress;
	BufferPool[i]->virtualAddr = (uint32 *) HWDEC_MMAP(0, size , PROT_READ | PROT_WRITE,
					    					MAP_SHARED, iMemAllocFd , params.busAddress);

	if ( BufferPool[i]->virtualAddr == MAP_FAILED )
	{
		/* MAP_FAILED equals to (uint32 *)(-1) , we shuold reset it to NULL , so the later judgement in 
		DeAllocate can determine the right state for virtualAddr */
		BufferPool[i]->virtualAddr = NULL;
		HWDEC_ERR("mmap linear buffer error , error = %d!\n" , errno);
		DeAllocate( (LBUF_HANDLER) i );
		return Ret_ErrUnkown;
	}

	BufferPool[i]->size = size;
//	HWDEC_MSG("Allocate Buffer : VA = 0x%08x PA = 0x%08x size = %d" , 
//					BufferPool[i]->virtualAddr , BufferPool[i]->bussAddr , BufferPool[i]->size);
	
	handler = (LBUF_HANDLER)i;

	return Ret_ErrNone;
}



void LinearBufferAllocator::DeAllocate(LBUF_HANDLER handler)
{
	uint32 index = (uint32)handler;
	LinearBufferCtrl *pCtrl = NULL;

	pCtrl = getLinearBufferCtrlByHandler(handler);

	if ( !pCtrl )
	{
		HWDEC_ERR("Coresponding buffer control structure doesnt available!\n");
		return ;
	}

	if ( pCtrl->virtualAddr )
	{
//		HWDEC_MSG("munmap VA = 08x%08x size = %d" , pCtrl->virtualAddr , pCtrl->size);
		HWDEC_MUNMAP( pCtrl->virtualAddr , pCtrl->size );
		pCtrl->virtualAddr = NULL;
	}

	if ( pCtrl->bussAddr != 0 )
	{
		HWDEC_IOCTL( iMemAllocFd , MEMALLOC_IOCSFREEBUFFER, &pCtrl->bussAddr);
		pCtrl->bussAddr = 0;
	}

	HWDEC_FREE(pCtrl);
	BufferPool[index] = NULL;
	iBufferNum--;

}



RET_STATUS LinearBufferAllocator::ReAllocate( uint32 size, LBUF_HANDLER handler , bool cpy)
{
	uint32 index = (uint32)handler;
	LinearBufferCtrl *pCtrl = NULL;

	pCtrl = getLinearBufferCtrlByHandler(handler);

	if ( !pCtrl )
	{
		HWDEC_ERR("Coresponding buffer control structure doesnt available!\n");
		return Ret_ErrNotFound;
	}
	else
	{
		LBUF_HANDLER newHandler;

		if ( Allocate( size, newHandler ) == Ret_ErrNone )
		{
			if ( cpy == true )
			{
				uint32 cpySize;
				LinearBufferCtrl *pNewCtrl;

				pNewCtrl = getLinearBufferCtrlByHandler( newHandler );

				/* This should never happen  */
				if ( !pNewCtrl )
				{
					HWDEC_ERR( " Cannot find entry of new allocate buffer!\n " );
					DeAllocate(newHandler);
					return Ret_ErrNotFound;
				}

				cpySize = ( (pCtrl->size > pNewCtrl->size)?pNewCtrl->size:pCtrl->size );
				HWDEC_MEM_CPY( pNewCtrl->virtualAddr , pCtrl->virtualAddr , cpySize);
			}

			/* Time to remove the old buffer */
			DeAllocate(handler);
			
			/* Swtich the content between the new ctrl and old one */
			BufferPool[handler] = BufferPool[newHandler];
			BufferPool[newHandler] = NULL;
		}
		else
		{
			HWDEC_ERR(" Allocate new buffer error \n");
			return Ret_ErrGeneral;
		}
	}

	return Ret_ErrNone;
}



uint32 LinearBufferAllocator::getBufferSize(LBUF_HANDLER handler)
{
	LinearBufferCtrl *pCtrl;

	pCtrl = getLinearBufferCtrlByHandler(handler);

	if ( pCtrl )
		return pCtrl->size;

	return 0;
}

uint8 * LinearBufferAllocator::getBufferVirtualAddr(LBUF_HANDLER handler)
{
	LinearBufferCtrl *pCtrl;

	pCtrl = getLinearBufferCtrlByHandler(handler);

	if ( pCtrl )
		return (uint8 *)(pCtrl->virtualAddr);

	return NULL;
}

uint32 LinearBufferAllocator::getBufferBussAddr(LBUF_HANDLER handler)
{
	LinearBufferCtrl *pCtrl;

	pCtrl = getLinearBufferCtrlByHandler(handler);

	if ( pCtrl )
		return pCtrl->bussAddr;

	return 0;
}

