
#include "simple_chain.h"
#include "hwdec_os_api.h"

simpleChain::simpleChain()
{
	chainHead = NULL;
	iChainEntryNum = 0;
	iSelector = NULL;
	iExplainer = NULL;
}

simpleChain::simpleChain( simpleChainSelectionInterface *selector , 
				simpleChainDataExplainerInterface *explainer)
{
	chainHead = NULL;
	iChainEntryNum = 0;
	iSelector = selector;
	iExplainer = explainer;
}


simpleChain::~simpleChain()
{
	EmptyChain();
}


void simpleChain::SetChainSelector( simpleChainSelectionInterface *selector  )
{
	iSelector = selector;
}

void simpleChain::SetDataExplainer(simpleChainDataExplainerInterface * explainer)
{
	iExplainer = explainer;
}


RET_STATUS simpleChain::ReplaceData( uint8 *data , uint32 dataLen , void *aParam)
{
	chainEntry *pEntry = chainHead;

	while(pEntry)
	{
		if ( iSelector->SelectionCallBack( (uint8 *)pEntry->data , pEntry->dataLen , aParam) )
		{
			if ( dataLen > pEntry->dataLen )
			{
				/* We need to realloc to get a bigger room to contain the data */
				void *pNewBuffer;

				HWDEC_MSG("Set data new len = %d , old = %d" , dataLen , pEntry->dataLen);

				if ( ( pNewBuffer = HWDEC_REALLOC( pEntry->data , dataLen ) ) == NULL )
				{
					HWDEC_ERR("ReplaceData no memory");
					return Ret_ErrNoMemory;
				}

				HWDEC_MSG("Realloc success . old addr : 0x%08x , new addr : 0x%08x", pEntry->data , pNewBuffer);
				pEntry->data = pNewBuffer;
			}
	
			HWDEC_MEM_CPY( pEntry->data , data , dataLen );
			pEntry->dataLen = dataLen;

			return Ret_ErrNone;
		}

		pEntry = pEntry->next;
	}

	return Ret_ErrNotFound;
}


RET_STATUS simpleChain:: AddToChainTail( uint8 *data , uint32 dataLen )
{
	chainEntry *pEntry;

	pEntry = AllocateChainEntry( data, dataLen);

	if ( !pEntry )
		return Ret_ErrNoResources;

	chainEntry **pp = &chainHead;

	while( *pp != NULL )
	{
		pp = & ((*pp)->next);
	}

	pEntry->next = NULL;
	*pp = pEntry;

	iChainEntryNum++;

	return Ret_ErrNone;
}


RET_STATUS simpleChain::AddToChainHead( uint8 * data, uint32 dataLen )
{
	chainEntry *pEntry;

	pEntry = AllocateChainEntry( data, dataLen);

	if ( !pEntry )
		return Ret_ErrNoResources;

	pEntry->next = chainHead;
	chainHead = pEntry;

	iChainEntryNum++;

	return Ret_ErrNone;
}


RET_STATUS simpleChain::GetData( uint8 *pData , uint32 *dataLen , void *aParam )
{
	chainEntry *pEntry = chainHead;
	RET_STATUS status = Ret_ErrNoResources;

	while( pEntry )
	{
		if ( iSelector->SelectionCallBack( (uint8 *)pEntry->data , pEntry->dataLen , aParam) )
			break;

		pEntry = pEntry->next;
	}

	if ( pEntry )
	{
		if ( pEntry->dataLen < *dataLen )
			*dataLen = pEntry->dataLen;

		/* Copy proper length */	
		HWDEC_MEM_CPY( pData , pEntry->data , *dataLen);

		status = Ret_ErrNone;
	}

	return status;
}


RET_STATUS simpleChain::DelFromChain( void *aParam )
{
	chainEntry **ppEntry = &chainHead;
	RET_STATUS status = Ret_ErrNoResources;
	
	while( *ppEntry )
	{
		if ( iSelector->SelectionCallBack( (uint8 *)(*ppEntry)->data , (*ppEntry)->dataLen , aParam ) )
			break;

		ppEntry = &((*ppEntry)->next);
	}

	/* We found a entry which matches the request */
	if ( *ppEntry )
	{
		/* The oreder is very important here , first , remove the selected entry from chain ,
		then free the resources */
		chainEntry *selected = *ppEntry;
		*ppEntry = ( *ppEntry )->next;

		if ( selected->data )
			HWDEC_FREE(selected->data);
		
		HWDEC_FREE(selected);

		iChainEntryNum--;
		status = Ret_ErrNone;
	}

	return status;
}


void simpleChain::EmptyChain()
{
	while( chainHead )
	{
		chainEntry *pEntry = chainHead;
		chainHead = chainHead->next;

		if( pEntry->data ) 
			HWDEC_FREE( pEntry->data );

		HWDEC_FREE(pEntry);
		iChainEntryNum--;
	}
}

void simpleChain::DisplayAllDataInfo()
{
	chainEntry *pEntry = chainHead;

	while( pEntry )
	{
		if ( iExplainer )
			iExplainer->ExplainCallBack( (uint8 *)pEntry->data , pEntry->dataLen);
		else
			HWDEC_ERR("Unkown data , size = %d",pEntry->dataLen);

		pEntry = pEntry->next;
	}
}

chainEntry * simpleChain::AllocateChainEntry( uint8 *data , uint32 dataLen )
{
	chainEntry *pEntry;
	
	pEntry = (chainEntry *)HWDEC_MALLOC( sizeof(chainEntry));

	if ( !pEntry )
	{
		HWDEC_ERR( "Cannot allocate memory for entry" );
		return NULL;
	}

	HWDEC_MEM_SET( (void *)pEntry , 0 , sizeof(pEntry));

	pEntry->data = HWDEC_MALLOC( dataLen );

	if ( !pEntry->data )
	{
		HWDEC_ERR("Cannot allocate memory for entry data");
		HWDEC_FREE(pEntry);
		return NULL;
	}
	
	pEntry->dataLen = dataLen;
	HWDEC_MEM_CPY( pEntry->data , data , dataLen );

	return pEntry;
}


