/*
**
** Licensed under the Apache License, Version 2.0 (the "License");
** you may not use this file except in compliance with the License.
** You may obtain a copy of the License at
**
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS,
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
** See the License for the specific language governing permissions and
** limitations under the License
*/

#define LOG_TAG "CameraHardware"
#include <utils/Log.h>

#include "CameraHardware.h"
#include <fcntl.h> 
#include <sys/mman.h>
#include <signal.h>

extern "C"{
#include<unistd.h>
#include <pthread.h>
#include <signal.h>
}
#include <cutils/properties.h>

namespace android {

wp<CameraHardwareInterface> CameraHardware::singleton;

CameraHardware::CameraHardware()
                  : mParameters(), 
                    mPreviewHeap(0),
                    mRawHeap(0),
                    mRecordingHeap(0),
                    mPreviewFrameSize(0),
                    mNotifyCb(0),
                    mDataCb(0),
                    mDataCbTimestamp(0),
                    mCallbackCookie(0),
                    mMsgEnabled(0),										
                    previewStopped(true),
                    recordStopped(true),
                    fd_video_preview(0),
                    fd_video_picture(0),
                    driver_used(false)
		    
{
   	char value[PROPERTY_VALUE_MAX];
	
    	property_get("camera.hardware.preview.width", value, "352");
    	mPreview_Width = atoi(value);
    	LOGD("CameraHardware::init:mPreview_Width %d",mPreview_Width);

    	property_get("camera.hardware.preview.height", value, "288");
    	mPreview_Height = atoi(value);
    	LOGD("CameraHardware::init:mPreview_Height %d",mPreview_Height);

    	property_get("camera.hardware.picture.width", value, "640");
    	mPicture_Width = atoi(value);
    	LOGD("CameraHardware::init:mPicture_Width %d",mPicture_Width);

    	property_get("camera.hardware.picture.height", value, "480");
    	mPicture_Height = atoi(value);
    	LOGD("CameraHardware::init:mPicture_Height %d",mPicture_Height);

    	property_get("camera.debug.showfps", value, "0");
    	mDebugFps = atoi(value);
    	LOGD_IF(mDebugFps, "showfps enabled");
	
    	initDefaultParameters();  
    	/* whether prop "debug.camera.showfps" is enabled or not */

}

void CameraHardware::initDefaultParameters()
{ 
    	CameraParameters p;

    	p.setPreviewSize(mPreview_Width, mPreview_Height);   
    	p.setPreviewFrameRate(30-15);
    	p.setPreviewFormat("yuv420sp" /*"yuv422sp"*/);    /* use 420sp ,picure more clear! */
    	p.set("jpeg-quality", "100");                     /* maximum quality */
    	//p.set("antibanding", "auto");                   /* flicker detection and elimination,too light*/
    	p.set("whitebalance", "daylight");
    	p.setPictureSize(mPicture_Width, mPicture_Height); 
	p.set("picture-size-values","640x480");
    	p.setPictureFormat("jpeg");
    	//p.set("orientation", /*"landscape"*/"portrait");
    	if (setParameters(p) != NO_ERROR) {
        	LOGE("Failed to set default parameters?!");
    	}
}

void CameraHardware::initHeapLocked()
{
	// Create raw heap.
	int picture_width, picture_height;
	mParameters.getPictureSize(&picture_width, &picture_height);
	LOGD("initHeapLocked: picture size=%dx%d", picture_width, picture_height);

	int preview_width, preview_height;
	mParameters.getPreviewSize(&preview_width, &preview_height);
    	LOGD("initHeapLocked: preview size=%dx%d", preview_width, preview_height);

   	// Note that we enforce yuv420 in setParameters().
    	int how_big = preview_width * preview_height * 2;

    	// If we are being reinitialized to the same size as before, no
	// work needs to be done.
	if (how_big == mPreviewFrameSize)
		return;
	if(mPreviewFrameSize != 0)
		release();
	
	mPreviewFrameSize = how_big;

	// Make a new mmap'ed heap that can be shared across processes. 
	// use code below to test with pmem
	mPreviewHeap = new MemoryHeapBase(mPreviewFrameSize);
    	// Make an IMemory for each frame so that we can reuse them in callbacks.
    	mPreviewBuffer = new MemoryBase(mPreviewHeap, 0, mPreviewFrameSize);

      	mRecordingHeap = new MemoryHeapBase(mPreviewFrameSize);
    	mRecordingBuffer = new MemoryBase(mRecordingHeap, 0, mPreviewFrameSize);
    
}

CameraHardware::~CameraHardware()
{  
	singleton.clear();
}

sp<IMemoryHeap> CameraHardware::getPreviewHeap() const
{
    return mPreviewHeap;
}

sp<IMemoryHeap> CameraHardware::getRawHeap() const
{
    return mRawHeap;
}

void CameraHardware::setCallbacks(notify_callback notify_cb,
                                      data_callback data_cb,
                                      data_callback_timestamp data_cb_timestamp,
                                      void* user)	
{
	Mutex::Autolock lock(mLock);
    	mNotifyCb = notify_cb;
    	mDataCb = data_cb;
    	mDataCbTimestamp = data_cb_timestamp;
    	mCallbackCookie = user;
}

void CameraHardware::enableMsgType(int32_t msgType)
{
	Mutex::Autolock lock(mLock);
	mMsgEnabled |= msgType;
}

void CameraHardware::disableMsgType(int32_t msgType)
{
	Mutex::Autolock lock(mLock);
    	mMsgEnabled &= ~msgType;
}

bool CameraHardware::msgTypeEnabled(int32_t msgType)
{
	Mutex::Autolock lock(mLock);
    return (mMsgEnabled & msgType);
}

static inline void* byteOffset(void* p, unsigned int offset)
{
    return (void*)((unsigned char*)p + offset);
}

static void showFPS(const char *tag)
{
	static int mFrameCount = 0;
	static int mLastFrameCount = 0;
    	static nsecs_t mLastFpsTime = 0;
    	static float mFps = 0;
    	mFrameCount++;
    	if (!(mFrameCount & 0x1F)) {
       	nsecs_t now = systemTime();
       	nsecs_t diff = now - mLastFpsTime;
        	mFps =  ((mFrameCount - mLastFrameCount) * float(s2ns(1))) / diff;
        	mLastFpsTime = now;
        	mLastFrameCount = mFrameCount;
        	LOGD("[%s] %d Frames, %f FPS", tag, mFrameCount, mFps);
    	}
}

#ifdef USE_OWNER_THREAD
 int CameraHardware::previewThread()
{	
	if(!previewStopped) {
		driver_used = true;
		{
			Mutex::Autolock lock(camLock);
			camera.GrabRawFrame(mPreviewHeap->getBase(),fd_video_preview);
		}
		
		if ((mMsgEnabled & CAMERA_MSG_PREVIEW_FRAME) ||(mMsgEnabled & CAMERA_MSG_VIDEO_FRAME)){	
			camera.rgb16TOyuv420((unsigned char *)mPreviewHeap->getBase(),(unsigned char *)mRecordingHeap->getBase());
			if ((mMsgEnabled & CAMERA_MSG_VIDEO_FRAME)) {
				nsecs_t timeStamp = systemTime(SYSTEM_TIME_MONOTONIC);
				mDataCbTimestamp(timeStamp, CAMERA_MSG_VIDEO_FRAME,mRecordingBuffer, mCallbackCookie);

			}
			
			// Notify the client of a new frame.
			// The image format in mPreviewBuffer is RGB,this is used for lcd display
			// But the PreviewFormat we set is yuv420sp,so we need convert rgb to yuv420.
	           	mDataCb(CAMERA_MSG_PREVIEW_FRAME, mRecordingBuffer, mCallbackCookie);
			
	       	if (mDebugFps) 
	              	showFPS("Preview");
		}
		driver_used = false;
		return 0;
    	}
	else{	
		LOGD("============previewThread Exit==========");
		return 1;	
    	}
}

int CameraHardware::beginPreviewThread(void *arg)	
{
	if(arg==0){
		LOGE("beginPreviewThread: arg=NULL! \n");
		return -1;
    	}
    	CameraHardware *c = (CameraHardware *)arg;
    	int ret =1;
    	while(1){	
		ret = c->previewThread();	        
	 	if(ret == 1){ 
			LOGD(".......... beginPreviewThread Exit .........\n");
			break;
	   	}
      	}
	return 0;
}
#else
// ---------------------------------------------------------------------------
int CameraHardware::previewThread()
{
	{
		Mutex::Autolock lock(camLock);
		camera.GrabRawFrame(mPreviewHeap->getBase(),fd_video_preview);
	}
		
	if ((mMsgEnabled & CAMERA_MSG_PREVIEW_FRAME) ||(mMsgEnabled & CAMERA_MSG_VIDEO_FRAME)){	
		camera.rgb16TOyuv420((unsigned char *)mPreviewHeap->getBase(),(unsigned char *)mRecordingHeap->getBase());
		if ((mMsgEnabled & CAMERA_MSG_VIDEO_FRAME)) {
			nsecs_t timeStamp = systemTime(SYSTEM_TIME_MONOTONIC);
			mDataCbTimestamp(timeStamp, CAMERA_MSG_VIDEO_FRAME,mRecordingBuffer, mCallbackCookie);

		}
			
		// Notify the client of a new frame.
		// The image format in mPreviewBuffer is RGB,this is used for lcd display
		// But the PreviewFormat we set is yuv420sp,so we need convert rgb to yuv420.
          	mDataCb(CAMERA_MSG_PREVIEW_FRAME, mRecordingBuffer, mCallbackCookie);
			
	      	if (mDebugFps) 
	             	showFPS("Preview");
	}	
    return NO_ERROR;
}
#endif

status_t CameraHardware::startPreview()
{   
    	int width, height,ret;
	/* To avoid the Zombie thread !*/
	signal(SIGCHLD,SIG_IGN);
    	Mutex::Autolock lock(mLock);
#ifdef USE_OWNER_THREAD
	if (previewStopped == false) {
#else
	if (mPreviewThread != 0){
#endif
        	//already running
       	LOGD("startPreview:  but already running");
        	return INVALID_OPERATION;
    	}
	
	fd_video_preview = camera.Open(mPreview_Width, mPreview_Height, PIXEL_FORMAT,1);
    	if (fd_video_preview < 0) {
		camera.Close(fd_video_preview); // should use this?
		LOGE("startPreview failed: cannot open device.");
        	return INVALID_OPERATION;
    	}

    	ret = camera.Init(fd_video_preview);
    	if(ret != 0){
		LOGE("startPreview:camera.Init error");
        	return UNKNOWN_ERROR;
    	}
		
    	ret = camera.StartStreaming(fd_video_preview);
    	if(ret != 0){
		LOGE("startPreview:camera.Init error");
        	return UNKNOWN_ERROR;
    	}
	
    	previewStopped = false;

	
  #ifdef USE_OWNER_THREAD 
   	if (createThread(beginPreviewThread, this) == false)
       	LOGD("---------create preview Thread error-------\n");
#else
	mPreviewThread = new PreviewThread(this);
#endif
	
    	return NO_ERROR;

}

void CameraHardware::stopPreview(){
	
	sp<PreviewThread> previewThread;
    
	{ // scope for the lock
		Mutex::Autolock lock(mLock);
		previewThread = mPreviewThread;
	}

	// don't hold the lock while waiting for the thread to quit
	if (previewThread != 0) {
		previewThread->requestExitAndWait();
	}

	Mutex::Autolock lock(mLock);
	mPreviewThread.clear();

	camera.Uninit(fd_video_preview);
	camera.StopStreaming(fd_video_preview);
	camera.Close(fd_video_preview);
	LOGD(".......... camera preview Exit .........\n");
}

bool CameraHardware::previewEnabled()	{
#ifdef USE_OWNER_THREAD
	return previewStopped == false;
#else
	return mPreviewThread != 0;
#endif
}

status_t CameraHardware::sendCommand(int32_t command, int32_t arg1,
                                         int32_t arg2){
	return 0;
}

status_t CameraHardware::startRecording(){
	{
		Mutex::Autolock lock(mLock);
		recordStopped = false;
	}
	return NO_ERROR;
}

void CameraHardware::stopRecording(){
	if (recordStopped == false){
		{ // scope for the lock
	        Mutex::Autolock lock(mLock);
	        recordStopped = true;
	   	}
		LOGD(".......... camera record Exit .........\n");
	}
	
}

bool CameraHardware::recordingEnabled(){
	return recordStopped == false;
}

void CameraHardware::releaseRecordingFrame(const sp<IMemory>& mem){
	return;
}

int CameraHardware::beginAutoFocusThread(void *cookie){
	CameraHardware *c = (CameraHardware *)cookie;
 	return c->autoFocusThread();
}

int CameraHardware::autoFocusThread()	{
    	if (mMsgEnabled & CAMERA_MSG_FOCUS)
       	mNotifyCb(CAMERA_MSG_FOCUS, true, 0, mCallbackCookie);
    	return NO_ERROR;
}

status_t CameraHardware::autoFocus(){  
	Mutex::Autolock lock(mLock);
    	if (createThread(beginAutoFocusThread, this) == false)
       	return UNKNOWN_ERROR;
    	return NO_ERROR;
}

status_t CameraHardware::cancelAutoFocus()	{
	return NO_ERROR;
}

int CameraHardware::pictureThread(){   
	LOGD("CameraHardware::pictureThread\n");
    	int ret;
    	int filesize;
    	int width, height;

	Mutex::Autolock lock(camLock);
    	if (mMsgEnabled & CAMERA_MSG_SHUTTER)
        	mNotifyCb(CAMERA_MSG_SHUTTER, 0, 0, mCallbackCookie);

	fd_video_picture = camera.Open(mPicture_Width, mPicture_Height, PIXEL_FORMAT,0);
	
    	if ( fd_video_picture < 0) {
		camera.Close(fd_video_picture);
        	LOGE("Cannot open device.\n");
        	return UNKNOWN_ERROR;
    	}

    	ret = camera.Init(fd_video_picture);

    	if(ret != 0){
		LOGE("pictureThread:camera.Init error\n");
       	return UNKNOWN_ERROR;
    	}

    	ret = camera.StartStreaming(fd_video_picture);

    	if(ret != 0){
		LOGE("pictureThread:camera.Init error\n");
       	return UNKNOWN_ERROR;
    	}
	
	filesize = camera.GrabJpegFrame(fd_video_picture);

	camera.Uninit(fd_video_picture);
   	camera.StopStreaming(fd_video_picture);
   	camera.Close(fd_video_picture);
	
	sp<MemoryHeapBase> mjpegPictureHeap = new MemoryHeapBase(filesize);
   	sp<MemoryBase> mjpegmemBase = new MemoryBase(mjpegPictureHeap, 0, filesize);

   	ret = camera.readjpeg(mjpegPictureHeap->base(),filesize);
   
   	if(ret != 0){
		LOGE("pictureThread:read jpeg error\n");
       	return UNKNOWN_ERROR;
    	}
   
     	if (mMsgEnabled & CAMERA_MSG_COMPRESSED_IMAGE) {
		mDataCb(CAMERA_MSG_COMPRESSED_IMAGE, mjpegmemBase, mCallbackCookie);
  	}
  		 
   	LOGD(".......... camera picture Exit .........\n");
   	return NO_ERROR;
   
}


int CameraHardware::beginPictureThread(void *cookie){
    	CameraHardware *c = (CameraHardware *)cookie;
    	return c->pictureThread();
}

status_t CameraHardware::takePicture(){    
    	if (createThread(beginPictureThread, this) == false){
		LOGD("takePicture:createThread faild\n");	
        	return -1;
    	}
    	return NO_ERROR;
}

status_t CameraHardware::cancelPicture(){

    return NO_ERROR;

}

status_t CameraHardware::dump(int fd, const Vector<String16>& args) const{
    return NO_ERROR;
}

status_t CameraHardware::setParameters(const CameraParameters& params){
    	Mutex::Autolock lock(mLock);

    	if (strcmp(params.getPictureFormat(), "jpeg") != 0) {
        	LOGE("Only jpeg still pictures are supported");
        	return -1;
    	}

    	int w, h;
    	int framerate;

    	params.getPreviewSize(&w, &h);
    	framerate = params.getPreviewFrameRate();
    	LOGD("## setParameters: w=%d h=%d framerate=%d", w, h, framerate);
	
/*
 * Dynamic set video width and height for recording in recorder and mms_video attach 2010.02.03
 * 
*/
    	w = (w < 176)? 176:w;
	h = (h < 144)? 144:h;

	/* Our original sensor image size is 640*480
	  * If we resizes the width to 480
	  * the height is resized to 360 according to the decimation module
	  */
	w = (w > 480)? 480:w;
	h = (h  > 360)? 360:h;
	
	mPreview_Width = w;
	mPreview_Height= h;
	
    	params.getPictureSize(&w, &h);
    	LOGD("requested picture size %d x %d", w, h);

	mPicture_Width = w;
	mPicture_Height= h;
	
	mParameters = params;
	
	initHeapLocked();

    return NO_ERROR;
}

CameraParameters CameraHardware::getParameters() const{
    	CameraParameters params;

    	{
    		Mutex::Autolock lock(mLock);
       	params = mParameters;
    	}

    return params;
}

void CameraHardware::release()
{
	LOGD("release:Clear the memory");
	mRawHeap.clear();
	mPreviewHeap.clear();
	mPreviewBuffer.clear();
	mRecordingHeap.clear();
	mRecordingBuffer.clear();
}


sp<CameraHardwareInterface> CameraHardware::createInstance()
{
    	if (singleton != 0) {
       	sp<CameraHardwareInterface> hardware = singleton.promote();
        	if (hardware != 0) {
            		return hardware;
        	}
    	}
    	sp<CameraHardwareInterface> hardware(new CameraHardware());
    	singleton = hardware;
    return hardware;
}

extern "C" sp<CameraHardwareInterface> openCameraHardware(){  
    return CameraHardware::createInstance();
}

}; // namespace android

