package android.net.ethernet;

import java.net.InetAddress;
import java.net.UnknownHostException;

import com.android.ethernettest.R;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.bluetooth.BluetoothHeadset;
import android.content.Context;
import android.content.Intent;
import android.net.ConnectivityManager;
import android.net.DhcpInfo;
import android.net.NetworkStateTracker;
import android.net.NetworkUtils;
import android.net.NetworkInfo.DetailedState;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.os.Message;
import android.os.Parcel;
import android.os.SystemProperties;
import android.util.*;
import android.os.ServiceManager;
import android.os.IBinder;
import android.os.SystemClock;

public class EthernetStateTracker extends Handler{

	private static final String TAG="EthernetStateTracker";
	
	private static final int EVENT_DHCP_START			= 0;
	private static final int EVENT_INTERFACE_CONFIGURATION_SUCCEEDED = 1;
	private static final int EVENT_INTERFACE_CONFIGURATION_FAILED	= 2;
	private static final int EVENT_HW_CONNECTED			= 3;
	private static final int EVENT_HW_DISCONNECTED			= 4;
	private static final int EVENT_HW_PHYCONNECTED			= 5;
	private static final int EVENT_MANUL_STOP				= 6;
	private static final int NOTIFY_ID				= 7;

	private EthernetManager mEM;
	
	private boolean mServiceStarted;

	private boolean mStackConnected;
	private boolean mHWConnected;
	private boolean mInterfaceStopped;
	private boolean mGetingIp;
	private DhcpHandler mDhcpTarget;
	private String mInterfaceName ;
	private DhcpInfo mDhcpInfo;
	private EthernetMonitor mMonitor;
	private String[] sDnsPropNames;
	private boolean mStartingDhcp;
	private NotificationManager mNotificationManager;
	private Notification mNotification;
	
	private Context mContext;
	
	public EthernetStateTracker(Context context,NotificationManager notify) {
	
		Log.i(TAG,"Starts...");

		mServiceStarted = true;

		mContext = context;

		if(EthernetNative.initEthernetNative() != 0 )
			Log.e(TAG,"Can not init ethernet device layers");
	
		HandlerThread dhcpThread = new HandlerThread("DHCP Handler Thread");

		mMonitor = new EthernetMonitor(this);
		
		dhcpThread.start();
		
		mDhcpTarget = new DhcpHandler(dhcpThread.getLooper(), this);
		
		mDhcpInfo = new DhcpInfo();

		mNotificationManager = notify;
		
	}

	public void Init(){
		getEthernetManager();
		Log.i(TAG,"Successed");
	}

	public void SetInterfaceState(int i){
		if(i == EthernetManager.ETH_STATE_ENABLED)
			mInterfaceStopped = false;
		else if(i == EthernetManager.ETH_STATE_DISABLED)
			mInterfaceStopped = true;
		postNotification();
	}

	public void SetInterfaceName(String i){
		mInterfaceName = i;
	}

	private void setHWState(boolean i){
		mHWConnected = i;
		postNotification();
	}

	private void setStackState(boolean i){
		mStackConnected = i;
		postNotification();
	}

	private void setGetingIPState(boolean i){
		mGetingIp = i;
		postNotification();
	}

	public boolean stopInterface(boolean suspend) {
		if (mEM != null) {
			EthernetDevInfo info = mEM.getSavedEthConfig();
			if (info != null && mEM.ethConfigured())
			{
				synchronized (mDhcpTarget) {
					
					Log.i(TAG, "stop dhcp and interface");
					
					mDhcpTarget.removeMessages(EVENT_DHCP_START);
					
					String ifname = info.getIfName();

					if (!NetworkUtils.stopDhcp(ifname)) {
						Log.e(TAG, "Could not stop DHCP");
					}
					
					NetworkUtils.resetConnections(ifname);
					NetworkUtils.removeDefaultRoute(ifname);

					mStartingDhcp = false;
				}
			}
		}

		return true;
	}

	private void getEthernetManager(){

		if (mEM == null) {
			
			IBinder b = ServiceManager.getService("ETH_SERVICE");
			IEthernetManager service = IEthernetManager.Stub.asInterface(b);
			mEM = new EthernetManager(service);
		}
	}

	private static int stringToIpAddr(String addrString) throws UnknownHostException {
		try {
			if (addrString == null)
				return 0;
				String[] parts = addrString.split("\\.");
				if (parts.length != 4) {
					throw new UnknownHostException(addrString);
			}

			int a = Integer.parseInt(parts[0]);
			int b = Integer.parseInt(parts[1]) <<  8;
			int c = Integer.parseInt(parts[2]) << 16;
			int d = Integer.parseInt(parts[3]) << 24;

			return a | b | c | d;
			
		} catch (NumberFormatException ex) {
		
			throw new UnknownHostException(addrString);
		}
	}

	private boolean configureInterface(EthernetDevInfo info) throws UnknownHostException {

		setGetingIPState(true);
		sDnsPropNames = new String[] {
						"dhcp." + mInterfaceName + ".dns1",
						"dhcp." + mInterfaceName + ".dns2"};

		if (info.getConnectMode().equals(EthernetDevInfo.ETH_CONN_MODE_DHCP)) {
			if (mStartingDhcp)
				return true;
			mStartingDhcp = true;
			Log.i(TAG, "trigger dhcp for device " + info.getIfName());
			mDhcpTarget.sendEmptyMessage(EVENT_DHCP_START);
		} else {
			int event;

			mStartingDhcp = false;

			mDhcpInfo.ipAddress =  stringToIpAddr(info.getIpAddress());
			mDhcpInfo.gateway = stringToIpAddr(info.getRouteAddr());
			mDhcpInfo.netmask = stringToIpAddr(info.getNetMask());
			mDhcpInfo.dns1 = stringToIpAddr(info.getDnsAddr());
			mDhcpInfo.dns2 = 0;

			Log.i(TAG, "set ip manually " + mDhcpInfo.toString());

			NetworkUtils.removeDefaultRoute(info.getIfName());

			if (NetworkUtils.configureInterface(info.getIfName(), mDhcpInfo)) {
				event = EVENT_INTERFACE_CONFIGURATION_SUCCEEDED;
				SystemProperties.set("net.dns1", info.getDnsAddr());
				SystemProperties.set("net."+info.getIfName()+".dns1", info.getDnsAddr());				
				Log.v(TAG, "Static IP configuration succeeded");
			} else {
				event = EVENT_INTERFACE_CONFIGURATION_FAILED;
				Log.v(TAG, "Static IP configuration failed");
			}
			
			this.sendEmptyMessage(event);
		}
		return true;
	}

	public boolean resetInterface()  throws UnknownHostException{
		/*
		 * This will guide us to enabled the enabled device
		 */
		 if(mEM == null){
			Log.e(TAG,"mEM is null,we will return.This should not happen!!!");
			return false;
		 }

		 if(!mHWConnected){
			Log.d(TAG,"The net wire is disconnect,we will not resetInterface");
			//Did not to throw a exception
			return true;
		 }

		 if(mInterfaceStopped){
			Log.d(TAG,"The network is stoped by user,we will not resetInterface");
			//Did not to throw a exception
			return true;
		 }
		 	

		EthernetDevInfo info = mEM.getSavedEthConfig();
		if (info != null && mEM.ethConfigured()){
			synchronized(this) {
				mInterfaceName = info.getIfName();
				Log.i(TAG, "reset device " + mInterfaceName);
				NetworkUtils.resetConnections(mInterfaceName);
				// Stop DHCP
				if (mDhcpTarget != null) {
					mDhcpTarget.removeMessages(EVENT_DHCP_START);
				}
						
				if (!NetworkUtils.stopDhcp(mInterfaceName)) {
					Log.e(TAG, "Could not stop DHCP");
				}
					
				configureInterface(info);
			}
		}
		return true;
	}
	
	public void StartPolling() {
		Log.i(TAG, "start polling");
		mMonitor.startMonitoring();
		EthernetNative.enableterface(mInterfaceName);
	}


	private void postNotification() {
		synchronized (this) {
			if (mNotificationManager != null) {
				int icon;
				CharSequence title = "Ethernet Status";
				CharSequence detail;
				if(mNotification == null) {
					mNotification = new Notification();
					mNotification.contentIntent = PendingIntent.getActivity(mContext, 0,
							new Intent(EthernetManager.NETWORK_STATE_CHANGED_ACTION), 0);
				}
				
				mNotification.when = System.currentTimeMillis();
				icon = R.drawable.connect_established;

				if(!mHWConnected){
					mNotification.icon = R.drawable.connect_no;
					detail = "Connect is unestablished,Net work cable is not plugined.";
				}else if(mInterfaceStopped){
					mNotification.icon = R.drawable.connect_established;
					detail = "Connect is established,But network disabled by user.";
				}else if(mGetingIp){
					mNotification.icon = R.drawable.connect_established;
					detail = "Connect is established,Waiting for IP address.";
				}else if(!mStackConnected){
					mNotification.icon = R.drawable.connect_established;
					detail = "Connect is established,But get IP address failed.";
				}else if (mStackConnected){
					EthernetDevInfo info = mEM.getSavedEthConfig();
					mNotification.icon = R.drawable.connect_creating;
					String ipaddr;
					if (info.getConnectMode().equals(EthernetDevInfo.ETH_CONN_MODE_DHCP)){
						String ipprop="dhcp."+mInterfaceName+".ipaddress";
						ipaddr = SystemProperties.get(ipprop);
					}else{
						 ipaddr = info.getIpAddress();
					}
					detail = "Ethernet is connected. IP address: " + ipaddr ;
				}else{
					mNotification.icon = R.drawable.connect_no;
					detail = "Unknow State.";
				}
				
				Log.i(TAG, "post event to notification manager "+detail);
				mNotification.setLatestEventInfo(mContext, title, detail,mNotification.contentIntent );
				mNotificationManager.notify(icon, mNotification);
			} else {
				Log.i(TAG, "notification manager is not up yet");
			}
		}

	}

	public void handleMessage(Message msg) {

		synchronized (this) {
			switch (msg.what) {
			case EVENT_INTERFACE_CONFIGURATION_SUCCEEDED:
				setGetingIPState(false);
				setStackState(true);
				Log.i(TAG, "received configured events, stack: " + mStackConnected + " HW " + mHWConnected );
//TOSEE					setDetailedState(DetailedState.CONNECTED);
			break;

			case EVENT_INTERFACE_CONFIGURATION_FAILED:
				setGetingIPState(false);
				setStackState(false);
			break;
				
			case EVENT_HW_CONNECTED:
//				setHWState(true);
//				Log.i(TAG, "received connected events, stack: " + mStackConnected + " HW " + mHWConnected );
//				mHWConnected = true;
//TOSEE					setDetailedState(DetailedState.CONNECTED);
			break;
			
			case EVENT_HW_DISCONNECTED:
				stopInterface(true);
				setHWState(false);
				Log.i(TAG, "received disconnected events, stack: " + mStackConnected + " HW " + mHWConnected );
//TOSEE				setDetailedState(DetailedState.DISCONNECTED);
			break;

			case EVENT_HW_PHYCONNECTED:
				Log.i(TAG, "interface up event, kick off connection request");
				setHWState(true);
				try {
					resetInterface();
				} catch (UnknownHostException e) {
					Log.e(TAG, "Wrong ethernet configuration");
				}
				
			break;
			}
		}
	}

	private class DhcpHandler extends Handler {
		private Handler mTrackerTarget;

		public DhcpHandler(Looper looper, Handler target) {
			super(looper);
			mTrackerTarget = target;
		}

		public void handleMessage(Message msg) {
			int event;

			switch (msg.what) {
				case EVENT_DHCP_START:
				synchronized (mDhcpTarget) {
					if (!mInterfaceStopped) {
						Log.d(TAG, "DhcpHandler: DHCP request started");
						if (NetworkUtils.runDhcp(mInterfaceName, mDhcpInfo)) {
							event = EVENT_INTERFACE_CONFIGURATION_SUCCEEDED;
							Log.v(TAG, "DhcpHandler: DHCP request succeeded");
						} else {
							event = EVENT_INTERFACE_CONFIGURATION_FAILED;
							Log.i(TAG, "DhcpHandler: DHCP request failed: " +
							NetworkUtils.getDhcpError());
						}
						if(mStartingDhcp)
							mTrackerTarget.sendEmptyMessage(event);
					} 
					mStartingDhcp = false;
				}
				break;
				}

		  }
	}

	public void notifyPhyConnected(String ifname) {
		Log.i(TAG, "report interface is up for " + ifname);
		synchronized(this) {
			if(ifname.equals(mInterfaceName))
				this.sendEmptyMessage(EVENT_HW_PHYCONNECTED);
		}

	}

	public void notifyStateChange(String ifname,DetailedState state) {
		Log.i(TAG, "report new state " + state.toString() + " on dev " + ifname);

		if (ifname.equals(mInterfaceName)) {

			Log.i(TAG, "update network state tracker");
			synchronized(this) {
				if (state.equals(DetailedState.CONNECTED) ) {
					this.sendEmptyMessage(EVENT_HW_CONNECTED);
				} else {
					this.sendEmptyMessage(EVENT_HW_DISCONNECTED);
				}

			}
		}
	}

}
